from __future__ import absolute_import

import equibel.PredicateTree.predicate
from equibel.PredicateTree.predicate import *

import equibel.PredicateTree.predicate_parser
from equibel.PredicateTree.predicate_parser import *



from equibel.simbool.proposition import Prop

__falseProp = Prop(False)
__trueProp = Prop(True)

def simplify_term(P):
    """negation-atom: ~True -> False; ~False -> True
    negation-elimination: ~~Prop -> Prop
    Unary op-elimination: (|Prop) -> Prop; (&Prop) -> Prop
    false-disjunction: False | Prop -> Prop
    true-disjunction: True | Prop -> True
    false-conjunction: False & Prop -> False
    true-conjunction: True & Prop -> Prop
    opposite-elimination-conjunction: ~Prop1 | Prop2 -> True if Prop1 == Prop2
    opposite-elmination-disjunction: ~Prop1 & Prop2 -> False if Prop1 == Prop2"""
    if P.atomic:
        return P
    
    if P.get_op() == '~': 
        sub = P.get_terms()[0]
        if sub.atomic:
            if sub.name == True:
                return __falseProp
            if sub.name == False:
                return __trueProp
        if sub.get_op() == '~':
            return sub.get_terms()[0]
        return P
    
    if len(P.get_terms()) == 1:
        return P.get_terms()[0]

    def breakdown(P):
        positive = set()
        negative = set()
        negative_with_negation = set()
        for e in P.get_terms():
            if e.is_positive():
                positive.add(e)
            else:
                assert(e.get_op() == '~')
                negative.add(e.get_terms()[0])
                negative_with_negation.add(e)
        return positive, negative, negative_with_negation
    
    if P.get_op() == '&':
        pos, neg, negn = breakdown(P)

        if __falseProp in pos:
            return __falseProp

        if len(pos & neg) > 0:
            return __falseProp
        
        if __trueProp in pos:
            if len(pos) > 1 or len(neg) > 0:
                pos.remove(__trueProp)
            else:
                return __trueProp
        terms = list(pos)+list(negn)
        if len(terms) == 1:
            return terms[0]
        return Prop('&', *terms)

    if P.get_op() == '|':
        pos, neg, negn = breakdown(P)

        if __trueProp in pos:
            return __trueProp

        if len(pos & neg) > 0:
            return __trueProp
        
        if __falseProp in pos:
            if len(pos) > 1 or len(neg) > 0:
                pos.remove(__falseProp)
            else:
                return __falseProp
        
        terms = list(pos)+list(negn)
        if len(terms) == 1:
            return terms[0]
        
        return Prop('|', *terms)
    
    assert(False)

def associative_collect(P):
    return __associative_collect(P)[0]

def __associative_collect(P, sym = None):
    if P.is_literal():
        return [P]

    if P.get_op() == '~':
        return [Prop('~', *__associative_collect(P.get_terms()[0]))]

    if sym != P.get_op():
        col = []
        for e in P.get_terms():
            col += __associative_collect(e, P.get_op())
        return [Prop(P.get_op(), *col)]
    else:
        col = []
        for e in P.get_terms():
            r = __associative_collect(e, sym)
            col += r
        return col
    
    assert(False)

def simplify_everywhere(P):
    res = __simplify_everywhere(P)
    res_old = res
    while True:
        res = __simplify_everywhere(res)
        if res_old == res:
            break
        res_old = res
    return res

def __simplify_everywhere(P):
    if P.atomic:
        return P
    
    simplified = simplify_term(P)

    if simplified.atomic:
        return simplified
    
    return Prop(simplified.get_op(), *[__simplify_everywhere(sub) for sub in simplified.get_terms()])

def push_neg(P):
    if P.is_literal():
        return P

    if P.get_op() in ['&', '|']:
        return Prop(P.get_op(), *[push_neg(sub) for sub in P.get_terms()])
    
    if P.get_op() == '~':
        term = P.get_terms()[0]
        if term.is_atomic():
            return S

        if term.get_op() == '&':
            return Prop('|', *[push_neg(~sub) for sub in term.get_terms()])
        if term.get_op() == '|':
            return Prop('&', *[push_neg(~sub) for sub in term.get_terms()])
        if term.get_op() == '~':
            return push_neg(term.get_terms()[0])
    
    assert(False)

def propagate_hypothesis(P, domain=set()):
    positive = set()
    negative = set()
    for e in domain:
        if e == __falseProp:
            return __falseProp
        if e.is_atomic():
            positive.add(e)
        else:
            negative.add(~e)
    if len(positive & negative) > 0:
        return __falseProp
    
    if P.is_literal():
        if simplify_term(~P) in domain:
            return __falseProp
        if P in domain:
            return __trueProp
        return P

    if P.get_op() in ['~']:
        return Prop('~', propagate_hypothesis(P.get_terms()[0], domain))

    if P.get_op() in ['&', '|']:
        new_terms = []
        old_terms = []
        new_domain = set()
        for e in P.get_terms():
            if simplify_term(~e) in domain:
                new_terms.append(__falseProp)
            else:
                if e in domain:
                    new_terms.append(__trueProp)
                else:
                    if e.is_literal():
                        new_terms.append(e)
                    else:
                        old_terms.append(e)

            if e.is_literal():
                if P.get_op() == '&':
                    new_domain.add(e)
                else:
                    new_domain.add(simplify_term(~e))

        for e in domain:
            new_domain.add(e)
        
        return Prop(P.get_op(), *(new_terms + [propagate_hypothesis(sub, new_domain) for sub in old_terms]))

    assert(False)

def factor_local(P):
    if P.is_literal():
        return P

    subterms = []
    for term in P.get_terms():
        if term.is_literal():
            if term.is_atomic():
                subterms.append([{term}, {term}, term])
            else:
                subterms.append([{simplify_term(~term)}, {term}, term])
        else:
            sub_p = set()
            sub_p_raw = set()
            for sub in term.get_terms():
                if sub.is_positive():
                    sub_p.add(sub)
                    sub_p_raw.add(sub)
                else:
                    sub_p.add(simplify_term(~sub))
                    sub_p_raw.add(sub)
            subterms.append([sub_p, sub_p_raw, term])
    
    counts = {}
    counts_raw = {}
    for s, r, _ in subterms:
        for ss in s:
            if ss not in counts:
                counts[ss] = 0
            counts[ss] += 1
        for sr in r:
            if sr not in counts_raw:
                counts_raw[sr] = 0
            counts_raw[sr] += 1
    
    # MODDED: Converted the dict_keys object returned by counts.keys() to a list, for Python3.
    sp = list(counts.keys())

    sp.sort(key = lambda x: counts[x], reverse=True)
    if counts[sp[0]] <= 2:
        # MODDED: Converted the dict_keys object returned by counts_raw.keys() to a list, for Python3.
        sp = list(counts_raw.keys())
        sp.sort(key = lambda x: counts_raw[x], reverse=True)
        if counts_raw[sp[0]] < 2:
            return P
        else:
            to_keep = [x[2] for x in subterms if sp[0] not in x[1]]
            to_factor = Prop(P.get_op(), *[x[2] for x in subterms if sp[0] in x[1]])
            return Prop(P.get_op(), (sp[0] & propagate_hypothesis(to_factor, {sp[0]})) | \
                                     (~sp[0] & propagate_hypothesis(to_factor, {simplify_term(~sp[0])})), *to_keep)
    
    to_keep = [x[2] for x in subterms if sp[0] not in x[0]]
    to_factor = Prop(P.get_op(), *[x[2] for x in subterms if sp[0] in x[0]])
    return Prop(P.get_op(), (sp[0] & propagate_hypothesis(to_factor, {sp[0]})) | \
                            (~sp[0] & propagate_hypothesis(to_factor, {simplify_term(~sp[0])})), *to_keep)

def factor_at_top(P):
    if P.is_literal():
        return P
    
    factored = factor_local(P)

    if factored is not P:
        return factored
    
    return Prop(P.get_op(), *[factor_at_top(sub) for sub in P.get_terms()])

def simplify_basic(P):
    res = associative_collect(P)
    res = simplify_everywhere(res)
    res = push_neg(res)
    res = associative_collect(res)
    return simplify_everywhere(res)

def simplify_pure(P):
    """Apply various strategies until reaching fixed point."""
    res = simplify_basic(P)
    
    old = res
    res = propagate_hypothesis(res)
    while True:
        res = associative_collect(res)
        res = simplify_everywhere(res)
        if old == res:
            res = factor_at_top(res)
            if res == old:
                break
            res = simplify_everywhere(res)
            res = associative_collect(res)
        old = res
        res = propagate_hypothesis(res)

    return res

def hypo(P, X):
    return simplify_pure((X & propagate_hypothesis(P, {X}))|((~X & propagate_hypothesis(P, {~X}))))

def hypo2(P, X, Y):
    return simplify_pure((X & Y & propagate_hypothesis(P, {X, Y})) | \
                    (~X & Y & propagate_hypothesis(P, {~X, Y})) | \
                    (X & ~Y & propagate_hypothesis(P, {X, ~Y})) | \
                    (~X & ~Y & propagate_hypothesis(P, {~X, ~Y})))

def hypo3(P, X, Y, Z):
    return simplify_pure((X & Y & Z & propagate_hypothesis(P, {X, Y, Z})) | \
                    (~X & Y & Z & propagate_hypothesis(P, {~X, Y, Z})) | \
                    (X & ~Y & Z & propagate_hypothesis(P, {X, ~Y, Z})) | \
                    (~X & ~Y & Z & propagate_hypothesis(P, {~X, ~Y, Z})) | \
                    (X & Y & ~Z & propagate_hypothesis(P, {X, Y, ~Z})) | \
                    (~X & Y & ~Z & propagate_hypothesis(P, {~X, Y, ~Z})) | \
                    (X & ~Y & ~Z & propagate_hypothesis(P, {X, ~Y, ~Z})) | \
                    (~X & ~Y & ~Z & propagate_hypothesis(P, {~X, ~Y, ~Z})))

def ultra_simplify(P, depth=2):
    """P should always be in simplified form. The depth
        parameter controls the size of the hypothesis, the
        running time is directly proportional to #atoms^depth, so
        use with precaution."""
    if P.size() < 6:
        return None
    
    # MODDED: Converted the dict_keys object returned by P.var_stats().keys() to a list, for Python3.
    vars_ = list(P.var_stats().keys())
    
    #try first order
    min_size = P.size()
    for var in vars_:
        c = hypo(P, var)
        if c.size() < min_size:
            min_size = c.size()
            candidate = c
    
    if min_size < P.size():
        return candidate
    
    if depth > 1:
        #try second order
        for i in range(len(vars_)):
            for j in range(i+1, len(vars_)):
                c = hypo2(P, vars_[i], vars_[j])
                if c.size() < min_size:
                    min_size = c.size()
                    candidate = c
        
        if min_size < P.size():
            return candidate
    
    if depth > 2:
        #try third order
        for i in range(len(vars_)):
            for j in range(i+1, len(vars_)):
                for k in range(j+1, len(vars_)):
                    c = hypo3(P, vars_[i], vars_[j], vars_[k])
                    if c.size() < min_size:
                        min_size = c.size()
                        candidate = c
        
        if min_size < P.size():
            return candidate
    
    # end by simplifying down the tree
    new_terms = []
    old_terms = []
    for sub in P.get_terms():
        x = ultra_simplify(sub, depth)
        if x is not None:
            new_terms.append(x)
        else:
            old_terms.append(sub)
    
    if len(new_terms) > 0:
        return simplify_pure(Prop(P.get_op(), *(new_terms+old_terms)))
    
    return None

def simplify(P, depth=2):
    res = simplify_pure(P)
    while True:
        res_new = ultra_simplify(res)
        if res_new is None:
            return res
        res = res_new



#    Copyright (C) 2014-2015 by
#    Paul Vicol <pvicol@sfu.ca>
#    All rights reserved.
#    BSD license.

from __future__ import absolute_import

import sys
import re

from equibel.graph import EquibelGraph
import equibel.parsers.FormulaParserSim as FormulaParserSim
import equibel.formatters.BCF_Formatter as BCF_Formatter


__all__ = ["parse_bcf"]


class FormatError(Exception):
    pass


# Can do one of two things here:
#    1) check that there is only one item in the tokens list,
#      and raise an exception if not, or
#    2) ignore any items after the first in the tokens list
#
# Using method (1), "t directed blah" is invalid, while using
# method (2), it is valid. How strict should the parser be?
#
# Since we want good error reporting for improperly formatted
# files, we will go with method (1).
#
# Philosophy: Things MUST be formatted correctly, but the
#           program will show you where any problems are.
#           This is more useful than ignoring an error, since
#           that error may be indicative of other problems
#           that the user did not notice.
def process_type_line(tokens, graph, line_num):
    if not tokens:
        raise FormatError("Empty type line on line {0}".format(line_num))
    if len(tokens) != 1:
        raise FormatError("Expected exactly 1 argument in type line on line {0}, got: \"{1}\""
                          .format(line_num, ", ".join(tokens)))

    graph_type = tokens[0].lower()

    if graph_type == 'directed':
        graph.to_directed()
    elif graph_type == 'undirected':
        graph.to_undirected()
    else:
        raise FormatError("unrecognized graph type on line {0}: \"{1}\""
                          .format(line_num, graph_type))


def process_node_line(tokens, graph, line_num):
    # There are two cases for valid tokens:
    #    1) The token can be an individual positive integer, like 3, or
    #    2) The token can be a range of the form x..y, where x and y are
    #      positive integers. To keep the parsing simple, whitespace is
    #      not allowed anywhere in the token, so "3.. 5" is invalid.

    if not tokens:
        raise FormatError("Empty node line on line {0}".format(line_num))

    for token in tokens:
        if '..' in token:
            m = re.match(r'([0-9]+)\.\.([0-9]+)', token)
            if not m:
                raise FormatError("Unrecognized range on line {0}: \"{1}\"".format(line_num, token))

            range_start = int(m.group(1))
            range_end   = int(m.group(2))

            if range_start > range_end:
                raise FormatError("End of range is smaller than start of range on line {0}: \"{1}\""
                                  .format(line_num, token))

            node_ids = range(range_start, range_end + 1)
            graph.add_nodes(node_ids)
        elif token.isdigit():
            node_id = int(token)
            graph.add_node(node_id)
        else:
            raise FormatError("Unrecognized node token on line {0}: \"{1}\"".format(line_num, token))


def process_atom_line(tokens, graph, line_num):
    if not tokens:
        raise FormatError("Empty atom line on line {0}".format(line_num))

    for token in tokens:
        if not re.match(r'[a-zA-Z][a-zA-Z0-9]*', token):
            raise FormatError("invalid atom name on line {0}: \"{1}\"".format(line_num, token))

        graph.add_atom(token)


def process_weight_line(tokens, graph, line_num):
    if not tokens:
        raise FormatError("Empty weight line on line {0}".format(line_num))
    if len(tokens) != 3:
        raise FormatError("Expected exactly 3 arguments in weight line on line {0}, got: \"{1}\""
                          .format(line_num, ", ".join(tokens)))

    node_str, atom_name, weight_str = tokens

    if not node_str.isdigit():
        raise FormatError("invalid node number on line {0}: \"{1}\"".format(line_num, node_str))
    if not re.match(r'[a-zA-Z][a-zA-Z0-9]*', atom_name):
        raise FormatError("invalid atom name on line {0}: \"{1}\"".format(line_num, atom_name))
    if not weight_str.isdigit():
        raise FormatError("invalid weight on line {0}: \"{1}\"".format(line_num, weight_str))

    node_id = int(node_str)
    weight  = int(weight_str)
    graph.add_weight(node_id, atom_name, weight)


def process_edge_line(tokens, graph, line_num):
    if not tokens:
        raise FormatError("Empty edge line on line {0}".format(line_num))
    if len(tokens) != 2:
        raise FormatError("Expected exactly 2 arguments in edge line on line {0}, got: \"{1}\""
                          .format(line_num, ", ".join(tokens)))

    from_node_str, to_node_str = tokens

    if not from_node_str.isdigit():
        raise FormatError("invalid node number on line {0}: \"{1}\"".format(line_num, from_node_str))
    if not to_node_str.isdigit():
        raise FormatError("invalid node number on line {0}: \"{1}\"".format(line_num, to_node_str))

    from_node_id = int(from_node_str)
    to_node_id = int(to_node_str)
    graph.add_edge(from_node_id, to_node_id)


def process_formula_line(node_str, formula_str, graph, line_num):
    if not node_str.isdigit():
        raise FormatError("Invalid node number on line {0}: \"{1}\"".format(line_num, node_str))
    if not formula_str:
        raise FormatError("No formula provided on line {0}".format(line_num))

    node_id = int(node_str)

    # TODO: May raise another exception.
    formula = FormulaParserSim.parse_formula(formula_str)
    graph.add_formula(node_id, formula)


process_line = {
    't': process_type_line,
    'n': process_node_line,
    'a': process_atom_line,
    'w': process_weight_line,
    'e': process_edge_line,
    'f': process_formula_line
}

def parse_bcf(filename):
    f = open(filename, 'r')

    # Creates the graph used to accumulate the information in the
    # BCF file. As each line is parsed, this graph is updated to
    # maintain the nodes, edges, atoms, weights, and formulas.
    graph = EquibelGraph()

    line_num = 1

    line = f.readline()

    while line:
        tokens = line.split()

        if tokens:
            line_type = tokens[0]
            if line_type in ['t', 'n', 'a', 'w', 'e']:
                process_line[line_type](tokens[1:], graph, line_num)
            elif line_type == 'f':
                formula_start_line = line_num
                if not tokens[1:]:
                    raise FormatError("Empty formula line on line {0}".format(formula_start_line))
                node_str = tokens[1]

                formula_str = " ".join(tokens[2:])
                while not balanced_parentheses(formula_str):
                    next_line = f.readline()
                    line_num += 1
                    if not next_line:
                        raise FormatError("Unbalanced parentheses on line {0}".format(line_num))

                    # The space ensures that tokens on successive lines don't get compacted into
                    # a single token, i.e. "p\n-q" becomes "p -q" instead of "p-q"
                    formula_str += " " + next_line.strip()

                process_formula_line(node_str, formula_str, graph, formula_start_line)
            elif line_type == 'c':
                pass
            else:
                raise FormatError("Unrecognized line type on line {0}: \"{1}\"".format(line_num, line_type))

        line_num += 1
        line = f.readline()

    f.close()
    return graph


# From StackOverflow
def balanced_parentheses(s):
    stack = []
    push_chars, pop_chars = "({[", ")}]"
    for c in s:
        if c in push_chars:
            stack.append(c)
        elif c in pop_chars:
            if not len(stack):
                return False
            else:
                stack_top = stack.pop()
                balancing_bracket = push_chars[pop_chars.index(c)]
                if stack_top != balancing_bracket:
                    return False

    return not len(stack)


if __name__ == '__main__':
    if len(sys.argv) < 2:
        print('usage: python BCF_Parser.py filename')
        sys.exit(1)

    filename = sys.argv[1]

    try:
        graph = parse_bcf(filename)
        print(BCF_Formatter.convert_to_bcf(graph))
    except FormatError as err:
        print(err)



from __future__ import absolute_import
from __future__ import print_function

import sys
import equibel.parsers.FormulaParserASP as FormulaParserASP

from equibel.simbool.proposition import *
from equibel.simbool.simplify import *

# TODO: See if this can be made cleaner.
def combine_formulas(models):
    disjunctions = dict()

    for model in models:
        conjunctions = dict()
        new_formula_predicates = model['new_formula']
        for new_formula_pred in new_formula_predicates:
            node_num, formula = extract_formula(new_formula_pred)
            if node_num in conjunctions:
                conjunctions[node_num] = simplify(conjunctions[node_num] & formula)
            else:
                conjunctions[node_num] = formula
        
        for node_num in conjunctions:
            conjunction = conjunctions[node_num]
            if node_num in disjunctions:
                disjunctions[node_num] = simplify(disjunctions[node_num] | conjunction)
            else:
                disjunctions[node_num] = conjunction
            
    # TODO: Some simplification.
    for node_num in disjunctions:
        disjunctions[node_num] = simplify(disjunctions[node_num])

    return disjunctions

# TODO: Add error checking?
def extract_formula(formula_term):
    arguments   = formula_term.args()
    node_id     = arguments[0]
    formula_str = str(arguments[1])
    formula    = FormulaParserASP.parse_asp_formula(formula_str)
    return (node_id, formula)



#    Copyright (C) 2014-2015 by
#    Paul Vicol <pvicol@sfu.ca>
#    All rights reserved.
#    MIT license.

from __future__ import absolute_import
from __future__ import print_function

def revise(K, phi):
    """
    """
    pass



from __future__ import absolute_import
from __future__ import print_function

import sys
import tempfile
import copy
import re

from subprocess import Popen, PIPE

import equibel.FormulaExtractor as FormulaExtractor
import equibel.formatters.ASP_Formatter as ASP_Formatter
import equibel.parsers.ASP_Parser as ASP_Parser

from equibel.simbool.simplify import simplify

CARDINALITY_TEMPLATE = "clingo equibel/eq_sets.lp equibel/cardinality_max.lp equibel/transitive.lp equibel/translate.lp {0} --opt-mode=optN --quiet=1,2 --verbose=0"
CONTAINMENT_TEMPLATE = "clingo equibel/eq_sets.lp equibel/transitive.lp equibel/translate.lp {0} 0 --heuristic=domain --enum-mode=domRec --verbose=0"

class UnsatisfiableError(Exception): pass

def run_one_shot_cardinality(filename):
    proc = Popen(CARDINALITY_TEMPLATE.format(filename), shell=True, stdout=PIPE, universal_newlines=True)
    return proc.stdout.read()

def run_one_shot_containment(filename):
    proc = Popen(CONTAINMENT_TEMPLATE.format(filename), shell=True, stdout=PIPE, universal_newlines=True)
    return proc.stdout.read()

def one_shot(graph, cardinality_maximal=False):
    # Returns a modified copy of the graph, incorporating the 
    # new formulas obtained from running the ASP solver.

    #temp_file = tempfile.NamedTemporaryFile(mode='w')
    temp_file = open('temp_asp_file', 'w')
    asp_str = ASP_Formatter.convert_to_asp(graph)
    #print("INPUT:\n", asp_str)

    temp_file.write(asp_str)
    temp_file.close()

    if cardinality_maximal:
        output = run_one_shot_cardinality(temp_file.name)
    else:
        output = run_one_shot_containment(temp_file.name)
    #print("OUTPUT:\n", output)

    models = ASP_Parser.parse_asp(output)
    node_formulas = FormulaExtractor.combined_formulas(models)
    new_graph = updated_graph(graph, node_formulas)
    return new_graph



ITERATE_CARD_TEMPLATE = "clingo equibel/eq_sets.lp equibel/cardinality_max.lp equibel/translate.lp {0} --opt-mode=optN --quiet=1,2 --verbose=0"
ITERATE_CONT_TEMPLATE = "clingo equibel/eq_sets.lp equibel/translate.lp {0} 0 --heuristic=domain --enum-mode=domRec --verbose=0"

def run_iterate_cardinality(filename):
    proc = Popen(ITERATE_CARD_TEMPLATE.format(filename), shell=True, stdout=PIPE, universal_newlines=True)
    return proc.stdout.read()

def run_iterate_containment(filename):
    #print("FILENAME: ", filename)
    proc = Popen(ITERATE_CONT_TEMPLATE.format(filename), shell=True, stdout=PIPE, universal_newlines=True)
    return proc.stdout.read()
    

def iterate(graph, num_iterations=1, cardinality_maximal=False):
    resultant_graph = graph

    for i in range(num_iterations):
        resultant_graph = iterate_once(resultant_graph, cardinality_maximal)

    return resultant_graph

def iterate_steady(graph, cardinality_maximal=False):
    old_graph = graph
    resultant_graph = iterate_once(old_graph, cardinality_maximal)

    while resultant_graph != old_graph:
        old_graph = resultant_graph
        resultant_graph = iterate_once(old_graph, cardinality_maximal)
    
    return resultant_graph


def iterate_once(graph, cardinality_maximal=False):
    #print("ITERATION")

    temp_file = open('temp_asp_file', 'w')
    asp_str = ASP_Formatter.convert_to_asp(graph)

    #print("Iteration input:\n", asp_str)
    temp_file.write(asp_str)
    temp_file.close()

    if cardinality_maximal:
        output = run_iterate_cardinality(temp_file.name)
    else:
        output = run_iterate_containment(temp_file.name)
    #print("Iteration output:\n", output)

    models = ASP_Parser.parse_asp(output)
    node_formulas = FormulaExtractor.combined_formulas(models)
    new_graph = updated_graph(graph, node_formulas)
    return new_graph


def updated_graph(graph, node_formulas):
    new_graph = copy.deepcopy(graph)
    for node_num in node_formulas:
        formula = node_formulas[node_num]

        # NOTE: This makes it so that true propositions are not added as new formulas.
        if not formula.is_true():
            new_graph.add_formula(node_num, formula)
    return new_graph



#    Copyright (C) 2014-2015 by
#    Paul Vicol <pvicol@sfu.ca>
#    All rights reserved.
#    BSD license.

import sys

import ply.lex
import ply.yacc

from equibel.simbool.proposition import *
from equibel.simbool.simplify import *

__all__ = ["parse_asp_formula"]

# --------------------------------------------------------------------
#                             LEXER
# --------------------------------------------------------------------

keywords = {"neg": "NEG", "and": "AND", "or": "OR", "implies": "IMPLIES", "iff": "IFF"}
tokens = (["IDENTIFIER", "LPAREN", "RPAREN", "COMMA"] + list(keywords.values()))

t_LPAREN = r'\('
t_RPAREN = r'\)'
t_COMMA  = r','

t_ignore = " \t\n"

def t_IDENTIFIER(t):
    r"[_a-zA-Z][_a-zA-Z0-9]*"
    t.type = keywords.get(t.value, "IDENTIFIER")
    return t

def t_NEWLINE(t):
    r"\n+"
    t.lexer.lineno += len(t.value)
    return t

def t_error(t):
    line = t.value.lstrip()
    i = line.find("\n")
    line = line if i == -1 else line[:i]
    raise ValueError("Syntax error, line {0}: {1}".format(t.lineno + 1, line))


# --------------------------------------------------------------------
#                             PARSER
# --------------------------------------------------------------------

def p_FORMULA(p):
    """FORMULA : ATOM
               | COMPOUND"""
    p[0] = p[1]

def p_ATOM(p):
    """ATOM : IDENTIFIER"""
    p[0] = Prop(p[1])

def p_COMPOUND(p):
    """COMPOUND : NEGATION
                | CONJUNCTION
                | DISJUNCTION
                | IMPLICATION
                | EQUIVALENCE"""
    p[0] = p[1]

def p_NEGATION(p):
    """NEGATION : NEG LPAREN FORMULA RPAREN"""
    p[0] = ~ p[3]

def p_CONJUNCTION(p):
    """CONJUNCTION : AND LPAREN FORMULA COMMA FORMULA RPAREN"""
    p[0] = p[3] & p[5]

def p_DISJUNCTION(p):
    """DISJUNCTION : OR LPAREN FORMULA COMMA FORMULA RPAREN"""
    p[0] = p[3] | p[5]

def p_IMPLICATION(p):
    """IMPLICATION : IMPLIES LPAREN FORMULA COMMA FORMULA RPAREN"""
    # Material implication using operator ">"
    p[0] = p[3] > p[5]

def p_EQUIVALENCE(p):
    """EQUIVALENCE : IFF LPAREN FORMULA COMMA FORMULA RPAREN"""
    p[0] = (p[3] > p[5]) & (p[5] > p[3])

def p_error(p):
    if p is None:
        raise ValueError("Unknown error")
    raise ValueError("Syntax error, line {0}: {1}".format(p.lineno + 1, p.type))

lexer = ply.lex.lex(debug=False)
parser = ply.yacc.yacc(debug=False, write_tables=False)

def parse_asp_formula(text):
    try:
        return parser.parse(text, lexer=lexer)
    except ValueError as err:
        print(err)
        return []


if __name__ == '__main__':
    if len(sys.argv) < 2:
        print('usage: python3 FormulaParserASP.py file')
        sys.exit(1)

    filename = sys.argv[1]
    f = open(filename, 'r')
    
    contents = f.read().strip()
    lines = contents.split("\n")
    for line in lines:
        formula = parse_asp_formula(line)
        print("Formula = {0}".format(repr(formula)))
        print("Simplified = {0}".format(repr(simplify(formula))))



import unittest
from equibel.simbool.proposition import *
from equibel.simbool.simplify import *

class TestSet(unittest.TestCase):
    def setUp(self):
        A = Prop('A')
        B = Prop('B')
        C = Prop('C')
        D = Prop('D')
        P = (C & ~D) | (D & ~C)
        T = Prop(True)
        F = Prop(False)

        self.cases = [T, F, A, ~T, ~F, ~A, 
                      A & F, F & A, A & T, T & A,
                      A & B, B & A, A | F, F | A,
                      T | A, A | T, A | B, B | A,
                      F | F, T | T, A | A,
                      F & F, T & T, A & A, 
                      P, ~~A, ~~P, ~~T, ~~F
                     ]

    def test_prop(self):
        A = Prop('A')
        B = Prop('B')
        C = Prop('C')
        D = Prop('D')
        P = (C & ~D) | (D & ~C)
        T = Prop(True)
        F = Prop(False)
        
        reprs = ['True', 'False', 'A', '~True', '~False', '~A', 
                'A & False', 'False & A', 'A & True', 'A & True', 
                'A & B', 'A & B', 'A | False', 'False | A', 
                'A | True', 'A | True', 'A | B', 'A | B', 
                '(|False)', '(|True)', '(|A)',
                '(&False)', '(&True)', '(&A)',
                '(~C & D) | (C & ~D)', '~~A',
                '~~((~C & D) | (C & ~D))', '~~True', '~~False',
               ]
        
        new_commute = [T, F, A, ~T, ~F, ~A, 
                      F & A, A & F, T & A, A & T,
                      B & A, A & B, F | A, A | F,
                      A | T, T | A, B | A, A | B,
                      F | F, T | T, A | A,
                      F & F, T & T, A & A,
                      (~C & D) | (~D & C), ~~A,
                     ~~((~C & D) | (~D & C)), ~~T, ~~F,
                     ]
        
        atomic = [True, True, True, False, False, False,
                  False, False, False, False,
                  False, False, False, False,
                  False, False, False, False,
                  False, False, False,
                  False, False, False,
                  False, False, 
                  False, False, False,
                 ]

        i = 0
        for expr in self.cases:
            self.assertEqual(repr(expr), reprs[i])
            self.assertEqual(expr, new_commute[i])
            self.assertEqual(expr.is_atomic(), atomic[i])
            i += 1

    def test_simplify_term(self):
        A = Prop('A')
        B = Prop('B')
        C = Prop('C')
        D = Prop('D')
        P = (C & ~D) | (D & ~C)
        T = Prop(True)
        F = Prop(False)
        
        simp = ['True', 'False', 'A', 'False', 'True', '~A', 
                'False', 'False', 'A', 'A', 
                'A & B', 'A & B', 'A', 'A', 
                'True', 'True', 'A | B', 'A | B', 
                'False', 'True', 'A',
                'False', 'True', 'A',
                '(~C & D) | (C & ~D)', 'A',
                '(~C & D) | (C & ~D)', 'True', 'False',
               ]
        
        new_commute = [T, F, A, F, T, ~A, 
                      F, F, A, A,
                      B & A, A & B, A, A,
                      T, T, B | A, A | B,
                      F, T, A,
                      F, T, A,
                      (~C & D) | (~D & C), A,
                      (~C & D) | (~D & C), T, F,
                     ]
        
        atomic = [True, True, True, True, True, False,
                  True, True, True, True,
                  False, False, True, True,
                  True, True, False, False,
                  True, True, True,
                  True, True, True,
                  False, True,
                  False, True, True,
                 ]

        i = 0
        for expr in self.cases:
            expr = simplify_term(expr)
            self.assertEqual(repr(expr), simp[i])
            self.assertEqual(expr, new_commute[i])
            self.assertEqual(expr.is_atomic(), atomic[i])
            i += 1

if __name__ == '__main__':
    unittest.main()



class Predicate:
    def __init__(self, name, children=[]):
        self.name = name
        self.children = children

    def __str__(self):
        return "{0}({1})".format(self.name, ",".join([str(child) for child in self.children]))



#    Copyright (C) 2014-2015 by
#    Paul Vicol <pvicol@sfu.ca>
#    All rights reserved.
#    BSD license.

from __future__ import absolute_import

import sys
from equibel.PredicateTree import predicate_parser

#TODO: Handle cases where the input is UNSATISFIABLE - now an error occurs somewhere else.

def parse_model(line):
    predicate_strs = line.split()
    predicates = [predicate_parser.parse_predicate(pred_str) for pred_str in predicate_strs]
    #print([str(pred) for pred in predicates])

    predicate_groups = dict()
    predicate_names = set(pred.name for pred in predicates)
    for pred_name in predicate_names:
        predicate_groups[pred_name] = [pred for pred in predicates if pred.name == pred_name]

    return predicate_groups
    

def parse_models(lines):
    ignore_lines = ['Optimization', 'OPTIMUM', 'Answer', 'SATISFIABLE']

    models = []
    for line in lines:
        if not line:
            continue
        if any(line.startswith(ignore) for ignore in ignore_lines):
            continue
        models.append(parse_model(line))
    return models

def parse_asp(text):
    lines = text.strip().split("\n")
    return parse_models(lines)

if __name__ == '__main__':
    if len(sys.argv) < 2:
        print('usage: python3 ASP_Parser.py filename')
        sys.exit(1)

    filename = sys.argv[1]

    f = open(filename, 'r')
    text = f.read()

    try:
        models = parse_asp(text)
        for model in models:
            print("Model")
            for pred_name in model:
                print("{0}: ".format(pred_name))
                for predicate in model[pred_name]:
                    print(predicate)
                print()
    except Exception as err:
        print(err)



#    Copyright (C) 2014-2015 by
#    Paul Vicol <pvicol@sfu.ca>
#    All rights reserved.
#    BSD license.

from __future__ import absolute_import
from __future__ import print_function

import sys
import pkg_resources
import copy
import platform

import equibel

if platform.system() == 'Linux':
    if platform.architecture()[0] == '64bit':
        if platform.dist()[0] in ['centos', 'CentOS']:
            import equibel.includes.linux.bit64.centos.gringo as gringo
            from equibel.includes.linux.bit64.centos.gringo import Control, Model, Fun
        else:
            import equibel.includes.linux.bit64.gringo as gringo
            from equibel.includes.linux.bit64.gringo import Control, Model, Fun
    elif platform.architecture()[0] == '32bit':
        import equibel.includes.linux.bit32.gringo as gringo
        from equibel.includes.linux.bit32.gringo import Control, Model, Fun
elif platform.system() == 'Darwin':
    import equibel.includes.mac.gringo as gringo
    from equibel.includes.mac.gringo import Control, Model, Fun

import equibel.FormulaExtractor as FormulaExtractor
import equibel.formatters.ASP_Formatter as ASP_Formatter
from equibel.simbool.proposition import Prop
from equibel.simbool.simplify import simplify


CONTAINMENT = 'containment'
CARDINALITY = 'cardinality'

EQ_SETS_FILE = pkg_resources.resource_filename('equibel', 'asp/eq_sets.lp')
TRANSITIVE_FILE = pkg_resources.resource_filename('equibel', 'asp/transitive.lp')
TRANSLATE_FILE = pkg_resources.resource_filename('equibel', 'asp/translate.lp')
EQ_EXPANDING_FILE = pkg_resources.resource_filename('equibel', 'asp/eq_expanding.lp')


class EqSolver(object):

    def __init__(self):
        self.optimal_models = set()
        self.optimal_values = None

    def find_models(self, asp_string, method=CONTAINMENT):
        #print("ASP STRING = {0}".format(asp_string))
        self.optimal_models = set()
        self.optimal_values = None

        ctl = gringo.Control()
        ctl.load(EQ_SETS_FILE)
        ctl.load(TRANSITIVE_FILE)

        self._configure_control(ctl, method)

        ctl.add('base', [], asp_string)
        ctl.ground([('base', [])])
        ctl.solve(on_model=self.capture_optimal_models)

        #print("SOLVER MODELS = {0}".format(self.optimal_models))
        print("NUM EQ SETS = {0}".format(len(self.optimal_models)))
        return self.optimal_models


    def one_shot_eq(self, asp_string, method=CONTAINMENT):
        models = self.find_models(asp_string, method)
        return self.extract_eq(models)

    
    def extract_eq(self, models):
        eq_dicts = []
        for model in models:
            eq_dict = dict()
            for term in model:
                if term.name() == 'eq':
                    atom = str(term.args()[0])
                    node1 = term.args()[1]
                    node2 = term.args()[2]
                    if node2 not in eq_dict:
                        eq_dict[node2] = dict()
                        eq_dict[node2][node1] = [atom]
                    else:
                        if node1 not in eq_dict[node2]:
                            eq_dict[node2][node1] = [atom]
                        else:
                            eq_dict[node2][node1].append(atom)
            eq_dicts.append(eq_dict)
        return eq_dicts
    

    def one_shot_dicts(self, asp_string, method=CONTAINMENT):
        models = self.find_models(asp_string, method)
        return self.model_dicts(models)


    def model_dicts(self, models):
        model_list = []

        for model in models:
            term_dict = dict()
            for term in model:
                name = term.name()
                if name in term_dict:
                    term_dict[name].append(term)
                else:
                    term_dict[name] = [term]
            model_list.append(term_dict)

        return model_list

    def _configure_control(self, ctl, method):
        if method == CONTAINMENT:
            ctl.conf.solve.opt_mode   = 'ignore'
            ctl.conf.solve.enum_mode  = 'domRec'
            ctl.conf.solver.heuristic = 'domain'
            ctl.conf.solve.models = 0
        elif method == CARDINALITY:
            ctl.conf.solve.opt_mode   = 'optN'
            ctl.conf.solve.enum_mode  = 'auto'
            ctl.conf.solver.heuristic = 'none'

    def capture_optimal_models(self, model):
        model_atoms = frozenset(model.atoms(Model.SHOWN))
        opt_values = model.optimization()

        #print("Opt val = {0}".format(self.optimal_values))

        if self.optimal_values is None:
            self.optimal_values = opt_values
            self.optimal_models.add(model_atoms)
        elif opt_values == self.optimal_values:
            self.optimal_models.add(model_atoms)
        elif opt_values < self.optimal_values:
            self.optimal_values = opt_values
            self.optimal_models.clear()
            self.optimal_models.add(model_atoms)


"""
# TEST TEST TEST
def completion(G, solving_method=CONTAINMENT):
    solver = EqSolver()
    models = solver.one_shot_dicts(ASP_Formatter.convert_to_asp(G), solving_method)
    #print("MODELS = {0}".format(models))
    node_formulas = FormulaExtractor.combine_formulas(models)

    R = copy.deepcopy(G)

    # TODO: Decide whether to always represent "multiple formulas" as a
    # single conjunction, or whether to keep them separate.
    for node_id in R.nodes():
        if node_id in node_formulas:
            new_formula = node_formulas[node_id]
            old_formula = conjunction(R.formulas(node_id))
            R.set_formulas(node_id, [simplify(old_formula & new_formula)])
            #R.set_formulas(node_id, [old_formula & new_formula])

    return R
"""


def conjunction(formulas):
    result = Prop(True)

    for formula in formulas:
        #print("FORMULA : {0}".format(formula))
        result &= formula

    """
    res = simplify(result)
    if debug:
        print("SIMPLIFIED CONJUNCTION = {0}".format(res))
    return res
    """
    return result


def disjunction(formulas):
    result = Prop(False)

    for formula in formulas:
        result |= formula

    #return simplify(result)
    return result

debug = False

def completion(G, solving_method=CONTAINMENT):
    #print("COMPLETION")
    solver = EqSolver()
    eq_dicts = solver.one_shot_eq(equibel.convert_to_asp(G), solving_method)
    if debug:
        print("EQ DICTS = {0}".format(eq_dicts))
        print("NUMBER OF EQ DICTS = {0}".format(len(eq_dicts)))
    return final_formulas(G, eq_dicts)


def final_formulas(G, eq_dicts):
    if debug:
        print("In final formulas...")
        print("-------------------------------")
    R = copy.deepcopy(G)
    for node in R.nodes():
        res = new_formula(node, G, eq_dicts)
        if debug:
            print("NODE {0} NEW FORMULAS= {1}".format(node, repr(res)))
        new_form = conjunction([form for form in G.formulas(node)] + [res])
        if debug:
            print("NODE {0} CONJUNCTION OF NEW FORMULAS= {1}".format(node, repr(new_form)))
        R.set_formulas(node, [simplify(new_form)])
        #R.set_formulas(node, [new_form])
    return R


def new_formula(node, G, eq_dicts):
    if debug:
        print("In new formulas...")
        print("-------------------------------")
    conjunctions = set()
    for eq_dict in eq_dicts:
        formulas = []
        if node in eq_dict:
            other_nodes = [curr_node for curr_node in G.nodes() if curr_node != node]
            if debug:
                print("\tUsing EQ DICT: {0}".format(eq_dict))
                print("\t-------------------------------")
            for other_node in other_nodes:
                if debug:
                    print("\t\tTranslating from {0} -> {1}".format(other_node, node))
                    print("\t\t-------------------------------")
                #eq_atoms = eq_dict[node][other_node]
                eq_atoms = eq_dict[node].get(other_node, [])
                if debug:
                    print("\t\t\tEQ atoms = {0}".format(eq_atoms))
                formulas.append(translate_formulas(G.formulas(other_node), eq_atoms))
                if debug:
                    print("\t\t\tFormulas = {0}".format(formulas))
            conjunctions.add(conjunction([form for forms in formulas for form in forms]))
            if debug:
                print("\tCONJUNCTIONS = {0}".format(conjunctions))

    return simplify(disjunction(conjunctions))


def translate_formulas(formulas, eq_atoms):
    if debug:
        print("\t\t\tIn translate formulas...")
        print("\t\t\t-------------------------------")

    translated_formulas = []
    for formula in formulas:
        translated_form = translate_formula(formula, eq_atoms)
        if debug:
            print("\t\t\t\tOriginal = {0}, Translated = {1}".format(repr(formula), repr(translated_form)))
        translated_formulas.append(translated_form)

    #print("translate_formulas({0}, {1}) = {2}".format(formulas, eq_atoms, translated_formulas))
    return translated_formulas


def translate_formula(formula, eq_atoms):
    if debug:
        print("\t\t\t\t\tTranslating {0} with {1}".format(repr(formula), eq_atoms))
    if formula.is_atomic():
        if formula.get_name() in eq_atoms:
            return formula
        else:
            return ~formula

    op = formula.get_op()
    terms = formula.get_terms()
    translated_terms = [translate_formula(term, eq_atoms) for term in terms]
    return equibel.Prop(op, *translated_terms)


# =================================================
# Functions that involve implicit graph topologies:
#   - Revision
#   - Contraction
#   - Merging (Projection and Consensus Based)
#   - Extrapolation
# =================================================

def revise(K, R):
    G = equibel.path_graph(2)

    if isinstance(K, str):
        G.add_formula(0, equibel.parse_infix_formula(K))
    elif isinstance(K, list):
        for belief in K:
            G.add_formula(0, equibel.parse_infix_formula(belief))

    if isinstance(R, str):
        G.add_formula(1, equibel.parse_infix_formula(R))
    elif isinstance(R, list):
        for belief in R:
            G.add_formula(1, equibel.parse_infix_formula(belief))

    S = completion(G)
    return S.formulas(1)


def contract(K, C):
    pass

def proj_merge(belief_bases, entailment_constraint=None):
    G = equibel.star_graph(len(belief_bases))

    if entailment_constraint:
        G.add_formula(0, entailment_constraint)

    for (i, belief_base) in enumerate(belief_bases, 1):
        G.add_formula(i, belief_base)

    S = completion(G)
    return S.formulas(0)
    

def con_merge(belief_bases, entailment_based_constraints=None, consistency_based_constraints=None):
    pass


if __name__ == '__main__':
    solver = EqSolver()

    G = equibel.path_graph(20)
    G.add_formula(0, "p & q")
    G.add_formula(1, "a & b & c & d & e & f & g & h & i & j & ~p")
    #eq_dicts = solver.one_shot_eq(equibel.convert_to_asp(G), method=CONTAINMENT)
    #print(eq_dicts)

    #print(eq_dicts)
    #print("NEW FORMULA {0}".format(new_formulas(0, G, eq_dicts)))

    """
    formula = equibel.parse_infix_formula("(p & q) | r")
    eq_atoms = ['q']
    translated = translate_formula(formula, eq_atoms)
    print(repr(translated))
    """

    R = completion(G)
    for node_id in R.nodes():
        print("Node {0}, formulas = {1}".format(node_id, R.formulas(node_id)))


    #filename = sys.argv[1]
    #f = file(filename, 'r')
    #final_models = solver.one_shot_dicts(equibel.convert_to_asp(G), method=CONTAINMENT)
    #print(final_models)




#    Copyright (C) 2014-2015 by
#    Paul Vicol <pvicol@sfu.ca>
#    All rights reserved.
#    BSD license.

from __future__ import absolute_import
from __future__ import print_function

import os
from cmd import Cmd
from subprocess import Popen, PIPE

from colorama import Fore, Style

import equibel

import equibel.formatters.ASP_Formatter as ASP_Formatter
import equibel.formatters.BCF_Formatter as BCF_Formatter

import equibel.parsers.CmdLineParser as CmdLineParser
import equibel.parsers.FormulaParserSim as FormulaParserSim
import equibel.parsers.BCF_Parser as BCF_Parser

from equibel.graph_manager import GraphManager
from equibel.graph import EquibelGraph

import equibel.solver as solver


class ArgumentError(Exception):
    pass


manager = GraphManager()
G = EquibelGraph()
manager.add('g', G)

solving_method = solver.CONTAINMENT


class EquibelPrompt(Cmd):

    def check_silencing_terminator(self, arg_str):
        """
        Checks if arg_str ends with a 'silencing' terminator, such as
        a semicolon. If it does, this function strips off the terminator
        and returns the modified string, as well as the boolean True to
        indicate that the output should be silenced. If it does not, this
        function returns the string unmodified, as well as the boolean
        False to indicate that the output should not be silenced.
        """
        verbose = True
        arg_str = arg_str.strip()
        if arg_str.endswith(';'):
            arg_str = arg_str[:-1]
            verbose = False
        return (arg_str, verbose)


    def default(self, line):
        """
        Default behaviour if the command prefix is not recognized.
        """
        line, verbose = self.check_silencing_terminator(line)
        if line.isdigit():
            if int(line) in manager.current_context.nodes():
                print(manager.current_context.node[int(line)])
        else:
            print("\tUnrecognized command! Type \"help\" for a list of commands.")


    def do_graphs(self, arg_str):
        """
        Usage: graphs

        Prints the names of all existing graphs, and shows which graph is
        the current context.
        """
        _, verbose = self.check_silencing_terminator(arg_str)
        if verbose:
            self.print_graphs()


    def print_graphs(self):
        """Prints the names of all existing graphs."""
        print()
        print("\t", end="")
        for graph_name in manager:
            if manager[graph_name] is manager.current_context:
                print(Fore.GREEN + "--{0}--".format(graph_name) + Fore.RESET, end="  ")
            else:
                print(graph_name, end="  ")
        print("\n")


    def do_create_graph(self, arg_str):
        """
        Usage: create_graph GRAPH_NAME

        Creates a new graph with the given name, and switches to the 
        context of the new graph.

        Example: create_graph g2
        """
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        args = arg_str.split()

        graph_name = args[0]
        R = EquibelGraph()
        manager.add(graph_name, R)
        manager.set_context(graph_name)

        self.prompt = "equibel ({0}) > ".format(graph_name)
        if verbose:
            self.print_graphs()


    def do_create_path(self, arg_str):
        """Creates a path graph."""
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        args = arg_str.split()
        graph_name = args[0]
        num_nodes_str = args[1]

        direction = None

        if len(args) > 2:
            direction = args[2]

        if not num_nodes_str.isdigit():
            raise ValueError("Error: create_path requires an integer argument!")
        else:
            num_nodes = int(num_nodes_str)

            if direction == "<->" or direction == None:
                R = equibel.path_graph(num_nodes)
            elif direction == "->":
                R = equibel.path_graph(num_nodes, directed=True)
            elif direction == "<-":
                R = equibel.path_graph(num_nodes, directed=True).reverse()
            else:
                raise ValueError('Error: The direction specifier must be one of "<-", "<->", or "->"')
            
            manager.add(graph_name, R)
            manager.set_context(graph_name)

            self.prompt = "equibel ({0}) > ".format(graph_name)
        if verbose:
            self.print_graphs()

    
    def do_create_star(self, arg_str):
        """Creates a star graph."""
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        args = arg_str.split()
        graph_name = args[0]
        num_nodes_str = args[1]

        if not num_nodes_str.isdigit():
            raise ValueError("Error: create_star requires an integer argument!")
        else:
            num_nodes = int(num_nodes_str)
            R = equibel.star_graph(num_nodes)
            manager.add(graph_name, R)
            manager.set_context(graph_name)
            self.prompt = "equibel ({0}) > ".format(graph_name)
        if verbose:
            self.print_graphs()


    def do_create_complete(self, arg_str):
        """Creates a complete graph."""
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        args = arg_str.split()
        graph_name = args[0]
        num_nodes_str = args[1]

        if not num_nodes_str.isdigit():
            raise ValueError("Error: create_complete requires an integer argument!")
        else:
            num_nodes = int(num_nodes_str)
            R = equibel.complete_graph(num_nodes)
            manager.add(graph_name, R)
            manager.set_context(graph_name)
            self.prompt = "equibel ({0}) > ".format(graph_name)
        if verbose:
            self.print_graphs()


    def do_use(self, arg_str):
        """
        Usage: use GRAPH_NAME

        Switches the context to the graph with the specified name.
            
        Example: use g2
        """
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        args = arg_str.split()
        graph_name = args[0]
        try:
            manager.set_context(graph_name)
            self.prompt = "equibel ({0}) > ".format(graph_name)
        except Exception as err:
            print(err)
        

    def do_nodes(self, arg_str):
        """
        Usage: nodes

        Prints the nodes in the current graph context.
        """
        _, verbose = self.check_silencing_terminator(arg_str)
        if verbose:
            self.print_nodes()


    def print_nodes(self):
        """Prints the nodes in the current context."""
        nodes = manager.current_context.nodes()

        if not nodes:
            print("\n\tnodes: []\n")
        else:
            print("\n\tnodes: {0}\n".format(nodes))
        

    def do_add_node(self, arg_str):
        """
        Usage: add_node NODE_NUM

        Adds a node to the current graph context.
        The node must be identified by an integer (for now).

        Example: add_node 1
        """
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        G = manager.current_context
        args = arg_str.split()
        if len(args) != 1:
            print("Incorrect number of arguments to add_node!")
        else:
            node_str = args[0]
            if not node_str.isdigit():
                raise ValueError("Error: add_node requires an integer argument!")
            else:
                node_id = int(node_str)
                G.add_node(node_id)
                if verbose:
                    self.print_nodes()


    def do_add_nodes(self, arg_str):
        """
        Usage: add_nodes NODE_LIST

        Adds all the nodes from a list to the current context.

        Examples:

          Add nodes 1, 3, 5, and 7:
             add_nodes [1, 3, 5, 7]
              
          Add a range of all nodes from 1 to 100, inclusive:
             add_nodes [1..100]
        """
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        graph = manager.current_context

        try:
            node_list = CmdLineParser.parse_list(arg_str)
            graph.add_nodes(node_list)

        except Exception as err:
            print(err)

        if verbose:
            self.print_nodes()

                    
    def do_remove_node(self, arg_str):
        """
        Usage: remove_node NODE_NUM

        Removes a node (if it exists) from the nodes set.

        Example: remove_node 2
        """
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        graph = manager.current_context

        args = arg_str.split()
        if len(args) != 1:
            print("Incorrect number of arguments to remove_node!")
        else:
            node_str = args[0]
            if not node_str.isdigit():
                print("remove_node requires an integer argument!")
            else:
                graph.remove_node(int(node_str))
                if verbose:
                    self.print_nodes()
    

    def do_remove_nodes(self, arg_str):
        """
        Usage: remove_nodes NODE_LIST

        Removes all the nodes in the given list from the nodes set.

        Examples:

          Remove nodes 1, 3, 5, and 7:
             remove_nodes [1, 3, 5, 7]
          
          Remove all nodes from 1 to 10, inclusive:
             remove_nodes [1..10]
        """
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        graph = manager.current_context
        
        nodes = CmdLineParser.parse_list(arg_str)
        # choose either "remove_nodes_from" or "remove_nodes"
        graph.remove_nodes_from(nodes)

        if verbose:
            self.print_nodes()


    def do_edges(self, arg_str):
        """
        Usage: edges

        Prints the existing edges.
        """
        _, verbose = self.check_silencing_terminator(arg_str)
        if verbose:
            self.print_edges()


    def print_edges(self):
        """Prints the existing edges."""
        G = manager.current_context
        edges = G.edges()
        directed = G.is_directed()

        if not edges:
            print("\n\tedges: []\n")
        else:
            if directed:
                arrow = '->'
            else:
                arrow = '<->'
            print("\n\tedges:")
            for edge in edges:
                print("\t\t{0} {1} {2}".format(edge[0], arrow, edge[1]))
            print()
        

    def do_add_edge(self, arg_str):
        """
        Usage: add_edge (START_NODE, END_NODE)

        Adds an edge to the edges set.
        
        Example: add_edge (1,2)
        """
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        G = manager.current_context

        try:
            (from_node_id, to_node_id) = CmdLineParser.parse_tuple(arg_str)
            G.add_edge(from_node_id, to_node_id)

            if verbose:
                self.print_edges()
        except Exception as err:
            print(err)


    def do_add_edges(self, arg_str):
        """
        Usage: add_edges EDGE_LIST

        Adds all the edges in the given list to the edges set.

        Example: add_edges [(1,2), (2,3), (3,4)]
        """
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        G = manager.current_context

        try:
            edge_list = CmdLineParser.parse_list(arg_str)
            G.add_edges(edge_list)
            
            if verbose:
                self.print_edges()
        except Exception as err:
            print(err)

        

    def do_remove_edge(self, arg_str):
        """
        Usage: remove_edge (START_NODE, END_NODE)

        Removes an edge (if it exists) from the edges set.
           
        Example: remove_edge (1,2)
        """
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        G = manager.current_context

        try:
            (from_node_id, to_node_id) = CmdLineParser.parse_tuple(arg_str)
            G.remove_edge(from_node_id, to_node_id)
            if verbose:
                self.print_edges()
        except Exception as err:
            print(err)


    # Directed/Undirected Functions
    #--------------------------------------------------------------------------------
    
    def do_directed(self, arg_str):
        """
        Usage: directed

        Makes the edges in the current context directed.
        """
        _, verbose = self.check_silencing_terminator(arg_str)
        G = manager.current_context
        directed_graph = G.to_directed()
        #G.to_directed()
        manager.update_context(directed_graph)
        if verbose:
            self.print_edges()

    def do_undirected(self, arg_str):
        """
        Usage: undirected

        Makes the edges in the current context undirected.
        """
        _, verbose = self.check_silencing_terminator(arg_str)
        G = manager.current_context
        undirected_graph = G.to_undirected()
        #G.to_undirected()
        manager.update_context(undirected_graph)
        if verbose:
            self.print_edges()


    # Atom Functions
    #--------------------------------------------------------------------------------

    def do_add_atom(self, arg_str):
        """
        Usage: add_atom ATOM_NAME
           
        Adds an atom to the alphabet of the current graph context.

        Examples: add_atom p
                  add_atom the_sky_is_blue
        """
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        G = manager.current_context

        args = arg_str.split()
        if len(args) != 1:
            print("Expected 1 argument to add_atom!")
        else:
            atom_str = args[0]
            G.add_atom(atom_str)
            
            if verbose:
                self.print_atoms()


    def do_add_atoms(self, arg_str):
        """
        Usage: add_atoms LIST_OF_ATOMS
           
        Adds all the atoms in the list to the alphabet of the current context.

        Examples: add_atoms [p, q, r, s]
                  add_atoms [blue_sky, green_grass]
        """
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        G = manager.current_context

        try:
            atom_list = CmdLineParser.parse_list(arg_str)
            G.add_atoms(atom_list)
            
            if verbose:
                self.print_atoms()
        except Exception as err:
            print(err)

    def do_remove_atom(self, arg_str):
        """
        Usage: remove_atom ATOM_NAME 

        Removes an atom from the alphabet of the current context.

        Example: remove_atom p
        """
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        G = manager.current_context

        args = arg_str.split()
        if len(args) != 1:
            print("Expected 1 argument to remove_atom!")
        else:
            atom_str = args[0]
            G.remove_atom(atom_str)
            if verbose:
                self.print_atoms()


    def do_atoms(self, arg_str):
        """
        Usage: atoms
           
        Prints the atoms in the alphabet of the current context.
        """
        _, verbose = self.check_silencing_terminator(arg_str)
        if verbose:
            self.print_atoms()

    def print_atoms(self):
        """Prints the atoms in the alphabet of the current context."""
        G = manager.current_context
        atoms = G.atoms()
        if not atoms:
            print("\n\tatoms: {}\n")
        else:
            print("\n\tatoms: {0}\n".format(atoms))
        

    # Weight Functions
    #--------------------------------------------------------------------------------

    # TODO: These are sort of temporary in their current form.
    def do_add_weight(self, arg_str):
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        args = arg_str.split()
        # DONE: Finish this function, with error checking.
        if len(args) != 3:
            raise ArgumentError("Expected 3 arguments to add_weight!\
                                \nUsage: add_weight NODE_NUM ATOM WEIGHT")

        node_str, atom, weight_str = args
        if not node_str.isdigit():
            raise ArgumentError("The node identifier must be an integer: \"{0}\"".format(node_str))
        if not weight_str.isdigit():
            raise ArgumentError("The weight must be an integer: \"{0}\"".format(weight_str))

        node_id = int(node_str)
        weight = int(weight_str)

        G = manager.current_context
        G.add_weight(node_id, atom, weight)

        if verbose:
            self.print_weights(node_id)


    def print_weights(self, node_id):
        weights = manager.current_context.atom_weights(node_id)

        print()
        print("\tnode {0}:".format(node_id))
        for atom in weights:
            print("\t\t{0}: {1}".format(atom, weights[atom]))
        print()


    def print_all_weights(self):
        G = manager.current_context
        
        print()
        for node_id in G.nodes():
            print("\tnode {0}:".format(node_id))
            weights = G.atom_weights(node_id)
            for atom in weights:
                print("\t\t{0}: {1}".format(atom, weights[atom]))
        print()


    def do_weights(self, arg_str):
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        args = arg_str.split()
        
        if len(args) == 0:
            if verbose:
                self.print_all_weights()
        elif len(args) == 1:
            node_str = args[0]
            if not node_str.isdigit():
                raise ArgumentError("The node identifier must be an integer: \"{0}\"".format(node_str))
            if verbose:
                self.print_weights(int(node_str))
        else:
            raise ArgumentError("Expected 0 or 1 argument to weights!\nusage: weights [NODE_NUM]")



    # Formula Functions
    #--------------------------------------------------------------------------------

    def do_add_formula(self, arg_str):
        """
        Usage: add_formula NODE_NUM FORMULA
            
        Adds a formula to the specified node.
        The formula must be specified in infix notation.

        Example: add_formula 1 (p & q) -> ~r
        """
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        #args = arg_str.split(maxsplit=1)
        # Python2 version of split does not accept keywork args, so we use:
        args = arg_str.split(None, 1)

        G = manager.current_context

        if len(args) != 2:
            raise ArgumentError("Expected exactly 2 arguments to add_formula.\nusage: add_formula NODE_NUM FORMULA")

        node_str, formula_str = args

        if not node_str.isdigit():
            raise ArgumentError("The node identifier must be an integer.")

        node_id = int(node_str)
        #formula = FormulaParserSim.parse_formula(formula_str)
        formula = equibel.parse_infix_formula(formula_str)
        G.add_formula(node_id, formula)

        if verbose:
            self.print_formulas(node_id)


    def do_remove_formula(self, arg_str):
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        #args = arg_str.split(maxsplit=1)
        # Python2 version of split does not accept keywork args, so we use:
        args = arg_str.split(None, 1)

        G = manager.current_context

        if len(args) != 2:
            raise ArgumentError("Expected exactly 2 arguments to add_formula.\nusage: add_formula NODE_NUM FORMULA")

        node_str, formula_str = args

        if not node_str.isdigit():
            raise ArgumentError("The node identifier must be an integer.")

        node_id = int(node_str)
        #formula = FormulaParserSim.parse_formula(formula_str)
        formula = equibel.parse_infix_formula(formula_str)

        G.remove_formula(node_id, formula)

        if verbose:
            self.print_formulas(node_id)
        
    
    def do_formulas(self, arg_str):
        """
        Usage: formulas [NODE_NUM]

        If a node is specified, prints the formulas at that node.
        If no node is specified, prints the formulas at each node.
           
        Example: formulas
                 formulas 1
        """
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        args = arg_str.split()
        
        if len(args) == 0:
            if verbose:
                self.print_all_formulas()
        elif len(args) == 1:
            node_str = args[0]
            if not node_str.isdigit():
                raise ArgumentError("the node identifier must be an integer: \"{0}\"".format(node_str))
            if verbose:
                self.print_formulas(int(node_str))
        else:
            raise ArgumentError("Expected 0 or 1 argument to formulas!\nusage: formulas [NODE_NUM]")


    def print_formulas(self, node_id):
        formulas = manager.current_context.formulas(node_id)

        print()
        print("\tnode {0}:".format(node_id))
        for formula in formulas:
            print("\t\t{0}".format(repr(formula)))
        print()


    def print_all_formulas(self):
        G = manager.current_context
        
        print()
        for node_id in G.nodes():
            print("\tnode {0}:".format(node_id))
            formulas = G.formulas(node_id)
            for formula in formulas:
                print("\t\t{0}".format(repr(formula)))
        print()
            

    # ASP Functions
    # -------------------------------------------------------------------------------

    def do_asp(self, arg_str):
        """
        Usage: asp

        Prints out the ASP code representing the current graph context.
        Used mainly for debugging purposes, so you can see the ASP code
        that is being passed to the solver.
        """
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        G = manager.current_context

        if verbose:
            print("\n" + ASP_Formatter.convert_to_asp(G))
        


    # Load/Store Functions
    # -------------------------------------------------------------------------------

    def do_load(self, arg_str):
        """Loads a graph from a BCF file into the current context (overwriting it)."""
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        args = arg_str.split()

        if len(args) != 1:
            raise ArgumentError("Expected exactly 1 argument to load().")

        filename = args[0]
        parsed_graph = BCF_Parser.parse_bcf(filename)
        manager.update_context(parsed_graph)

        if verbose:
            print("\n\tgraph successfully loaded from \"{0}\"\n".format(filename))
        

    # DONE: Change the output format from ASP to BCF (after completing BCF_Formatter.py).
    def do_store(self, arg_str):
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        args = arg_str.split()

        graph = manager.current_context
        graph_name = manager.current_context_name

        if len(args) != 1:
            raise ArgumentError("Expected exactly 1 argument to store().")

        filename = args[0]
        if os.path.isfile(filename):
            confirm_overwrite = self.yes_or_no(("\tFile \"{0}\" already exists. Do you want to overwrite it? (y/n) ")
                                         .format(filename))

            if not confirm_overwrite:
                print("\n\t\"{0}\" left unchanged.\n".format(filename))
                return

        f = open(filename, 'w')
        f.write(BCF_Formatter.convert_to_bcf(graph))
        if verbose:
            print("\n\tSuccessfully saved graph {0} to {1}.\n".format(graph_name, filename))
        

    def yes_or_no(self, question):
        yes_answers = ['yes', 'y']
        no_answers  = ['no', 'n']

        answer = input(question)
        while answer not in (yes_answers + no_answers):
            print("\t\"{0}\" is not a valid response.".format(answer))
            answer = input(question)

        if answer in yes_answers:
            return True
        return False 




    # Belief Change Operations
    # -------------------------------------------------------------------------------
    # TODO: Needs checking for EquibelGraph
    def do_one_shot(self, arg_str):
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        graph = manager.current_context

        # new_graph = SolverInterface.one_shot(graph, cardinality_maximal)
        new_graph = solver.completion(graph, solving_method)
        manager.update_context(new_graph)

        if verbose:
            print("\n\tOne-shot belief change completed:")
            print("\t---------------------------------")
            self.print_all_formulas()


    # TODO: Needs checking for EquibelGraph
    def do_cardinality(self, arg_str):
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        # cardinality_maximal = True
        solving_method = solver.CARDINALITY
        if verbose:
            print("\n\tNow using cardinality-maximal EQ sets.\n")


    # TODO: Needs checking for EquibelGraph
    def do_containment(self, arg_str):
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        # cardinality_maximal = False
        solving_method = solver.CONTAINMENT
        if verbose:
            print("\n\tNow using containment-maximal EQ sets.\n")


    # TODO: Needs checking for EquibelGraph
    def do_iterate(self, arg_str):
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        args = arg_str.split()

        if len(args) == 0:
            num_iterations = 1
        elif len(args) == 1:
            num_str = args[0]
            if not num_str.isdigit():
                raise ArgumentError("The number of iterations must be an integer: \"{0}\"".format(num_str))
            num_iterations = int(num_str)
        else:
            raise ArgumentError("Expected 0 or 1 argument to iterate()!\ Usage: iterate [NUM_ITERATIONS]")

        graph = manager.current_context
        new_graph = solver.iterate(graph, num_iterations, solver.CONTAINMENT)
        manager.update_context(new_graph)

        if verbose:
            if num_iterations == 1:
                print("\n\t1 iteration completed.\n")
            else:
                print("\n\t{0} iterations completed.\n".format(num_iterations))


    # TODO: Needs checking for EquibelGraph
    def do_e_iterate(self, arg_str):
        arg_str, verbose = self.check_silencing_terminator(arg_str)
        args = arg_str.split()

        if len(args) == 0:
            num_iterations = 1
        elif len(args) == 1:
            num_str = args[0]
            if not num_str.isdigit():
                raise ArgumentError("The number of iterations must be an integer: \"{0}\"".format(num_str))
            num_iterations = int(num_str)
        else:
            raise ArgumentError("Expected 0 or 1 argument to iterate()!\ Usage: iterate [NUM_ITERATIONS]")

        graph = manager.current_context
        #new_graph = solver.expanding_iteration(graph, num_iterations, solver.CONTAINMENT)
        solver.expanding_iteration(graph, num_iterations, solver.CONTAINMENT)
        #manager.update_context(new_graph)

        if verbose:
            if num_iterations == 1:
                print("\n\t1 iteration completed.\n")
            else:
                print("\n\t{0} iterations completed.\n".format(num_iterations))


    # Shell Functions -- (To help locate files to load.)
    # -------------------------------------------------------------------------------

    def do_shell(self, arg_str):
        proc = Popen(arg_str, shell=True, stdout=PIPE, universal_newlines=True)
        for line in proc.stdout:
            line = line.strip()
            print(line)

    
    def do_quit(self, args):
        """Quits the program."""
        print("Bye!")
        raise SystemExit


def cli():
    print("Equibel Version 0.8.8 (Alpha)")

    print(Fore.GREEN + Style.BRIGHT)

    print("""
                                       ..                    ..
                                       ||                    ||
   ______     _______   ..      ..  ** ||_____      ______   || **
 //`````\\\\   //`````\\\\  ||      ||  || ||````\\\\   //`````\\\\  || ||
||_______|| ||       || ||      ||  || ||     || ||_______|| || ||
||````````  ||       || ||      ||  || ||     || ||````````  || ||
 \\\\______    \\\\_____//|  \\\\____//|  || ||____//   \\\\______   || ||
  ```````     ```````||   `````` \\\\ `` ```````     ```````   `` ``
                     ||
                      \\\\
                       ``
    """)

    print(Fore.RESET + Style.RESET_ALL)

    prompt = EquibelPrompt(completekey='tab')
    prompt.prompt = "equibel (g) > "

    while True:
        try:
            prompt.cmdloop()
        except Exception as err:
            print(err)


if __name__ == '__main__':
    cli()



"""Manage multiple graphs during an interactive session.

The GraphManager class keeps track of graph objects by 
associating them with names. This makes it easy to manage 
multiple graphs w.r.t switching between them and maintaining 
which is the current interactive context.
"""

#    Copyright (C) 2014-2015 by
#    Paul Vicol <pvicol@sfu.ca>
#    All rights reserved.
#    MIT license.

class GraphManagerError(Exception):
    pass


class GraphManager:
    """
    The GraphManager class is used to manage multiple
    graphs during an interactive session, and to keep
    track of which graph is the current context for
    commands issued at the prompt.

    Example
    -------
    Create two graphs:

    >>> g = Graph()
    >>> g2 = Graph()

    Create a GraphManager and add the two graphs to it:

    >>> manager = GraphManager()
    >>> manager.add('g', g)
    >>> manager.add('g2', g2)

    Retrieve the name of the current context. Note that
    the first graph added to the manager (graph 'g' above)
    automatically becomes the current context:

    >>> manager.current_context_name
    'g'

    Retrieve the current context graph, perform an operation
    on it to obtain a modified graph, and update the current
    context to reflect the changes:

    >>> context_graph = manager.current_context
    >>> completion_graph = equibel.completion(context_graph)
    >>> manager.update_context(completion_graph)
    """
    def __init__(self):
        self.graphs = dict()
        self.current_context = None
        self.current_context_name = None


    def __iter__(self):
        """
        Returns an iterator over the graphs.
        """
        return iter(self.graphs)


    def __getitem__(self, key):
        """
        Finds a graph by name. If the graph is not in
        the graph manager, returns None.

        Example
        -------
        >>> g2 = Graph()
        >>> manager = GraphManager()
        >>> manager.add('g2', g2)

        At any time, the graph can be accessed by:

        >>> manager['g2']
        """
        return self.graphs.get(key, None)


    def add(self, graph_name, graph):
        """
        Adds a graph object to this manager. The given
        graph name is used as the key to access the graph
        object later.

        Example
        -------
        Create a graph and a manager, and add the graph to
        the manager:

        >>> g2 = Graph()
        >>> manager = GraphManager()
        >>> manager.add('g2', g2)
        """
        self.graphs[graph_name] = graph
        # The first graph added to the manager becomes the
        # current context automatically.
        if not self.current_context:
            self.current_context = graph
            self.current_context_name = graph_name


    def update_context(self, graph):
        """
        Overwrites the graph in the current context with
        the given graph (to reflect changes to the context).
        That is, this associates the current context name
        with a new graph.

        Example
        -------
        >>> manager = GraphManager()
        >>> manager.add('g', Graph())

        Retrieve graph 'g' and perform an operation on it,
        creating a new graph:

        >>> working_graph = manager['g']
        >>> updated_graph = equibel.completion(working_graph)

        Update the context, associating the name 'g' with
        the updated_graph, and losing the reference to the
        working_graph:

        >>> manager.update_context(updated_graph)
        """
        self.graphs[self.current_context_name] = graph
        self.current_context = graph


    def remove(self, graph_name):
        """
        Removes a graph from the graph manager. Note that
        if the manager contains the only reference to the
        graph in question, then the graph will be lost.

        Example
        -------
        >>> manager.remove('g2')
        """
        if graph_name not in self.graphs:
            raise GraphManagerError("Graph \"{0}\" does not exist".format(graph_name))
        del self.graphs[graph_name]


    def set_context(self, graph_name):
        """
        Sets the current context to the graph identified by
        the given name.

        Example
        -------
        >>> manager = GraphManager()
        >>> manager.add('g', Graph())
        >>> manager.add('g2', Graph())
        >>> manager.current_context_name
        'g'
        >>> manager.set_context('g2')
        >>> manager.current_context_name
        'g2'
        """
        if graph_name not in self.graphs:
            raise GraphManagerError("Graph \"{0}\" does not exist".format(graph_name))
        self.current_context = self.graphs[graph_name]
        self.current_context_name = graph_name



from __future__ import absolute_import

from equibel.parsers.ASP_Parser import *
from equibel.parsers.FormulaParserASP import *
from equibel.parsers.FormulaParserSim import *
from equibel.parsers.formula_parser_infix import *
#from equibel.parsers.BCF_Parser import *



from __future__ import print_function
from functools import reduce
import operator

class Prop:
    def __init__(self, *args):
        """Proposition constructor.
        Prop(True), Prop(False) constructs the True and False atoms.
        Prop(name) constructs an atomic proposition.
        Prop('&', p1, p2, ...) constructs the conjunction of propositions p1, p2, ...
        Prop('|', p1, p2, ...) constructs the disjunction of propositions p1, p2, ...
        Prop('~', p) constructs the negation of proposition p.
        
        Built-in rewritings:
            redundance-elimination-union: Prop1 | Prop2 -> (|Prop1) if Prop1 == Prop2
            redundance-elimination-intersection: Prop1 & Prop2 -> (&Prop1) if Prop1 == Prop2"""
        if len(args) == 0:
            raise NameError("Malformed proposition.")
        
        if len(args) == 1:
            self.name = args[0]
            self.atomic = True

        if len(args) > 1:
            if args[0] not in ['&', '|', '~'] or \
                args[0] == '~' and len(args)>2:
                raise NameError("Malformed proposition: "+str(args)) 
            self.oper = args[0]
            self.terms = frozenset(args[1:])
            self.atomic = False

    # My function to handle return the names of True and False propositions as the
    # lowercase strings 'true' and 'false'
    def get_name(self):
        if self.atomic:
            if self.name == True:
                return 'true'
            elif self.name == False:
                return 'false'
            return self.name
        else:
            return None
        
    def is_true(self):
        return self.atomic and self.name == True

    # My function to get all the atom names.
    def get_atoms(self):
        if self.atomic:
            return [self.get_name()]
        return reduce(operator.add, [term.get_atoms() for term in self.terms])

    def is_atomic(self):
        return self.atomic
    
    def is_positive(self):
        return self.atomic or (self.oper in ['&', '|'])

    def is_negative(self):
        return (not self.atomic) and (self.oper == '~')

    def is_literal(self):
        return self.atomic or \
                self.oper == '~' and [x.atomic for x in self.terms][0]
    
    def __eq__(self, other):
        """Syntactic equality testing. Fully commutative but not associative. 
        If True, then the two propositions are equal. If False, unsure."""
        if self.atomic:
            if other.atomic:
                return self.name == other.name
            return False
        if other.atomic:
            return False
        if self.oper == self.oper:
            return self.terms == other.terms
        return False
    
    def __ne__(self, other):
        return not (self == other)
    
    def __hash__(self):
        if '_hash' in self.__dict__:
            return self._hash
        if self.atomic:
            self._hash = hash(self.name)
            return self._hash
        self._hash = hash(hash(self.oper)^hash(self.terms))
        return self._hash
    
    def __repr__(self):
        def no_parenthesing(s):
            return s.atomic or s.oper == '~'
        
        if self.atomic:
            return str(self.get_name())
        
        oper = self.oper
        if oper == '~':
            a = [x for x in self.terms][0]
            if no_parenthesing(a):
                return oper+repr(a)
            return oper+'('+repr(a)+')'
        
        if oper in ['&', '|']:
            subs = []
            for e in self.terms:
                if no_parenthesing(e):
                    subs.append(repr(e))
                else:
                    subs.append('('+repr(e)+')')
            assert(len(subs) > 0)
            if len(subs) == 1:
                return '('+oper+subs[0]+')'
            return (' '+oper+' ').join(subs)
        
        assert(False)

    def __and__(self, other):
        return Prop('&', self, other)

    def __rand__(self, other):
        return other.__and__(self)

    def __or__(self, other):
        return Prop('|', self, other)

    def __ror__(self, other):
        return other.__ror__(self)

    def __invert__(self):
        return Prop('~', self)

    def __gt__(self, other):
        """material implication."""
        return other | (~self)
    
    def __sub__(self, other):
        return self & ~other

    def __rsub__(self, other):
        return other.__sub__(self)
    
    def __pformat(self, indentation=0, space=' '):
        if self.is_literal():
            if self.is_positive():
                return str(self.name)
            return '~'+str(self.get_terms()[0].name)
        
        oper = self.oper
        if oper == '~':
            a = [x for x in self.terms][0]
            return oper+' '+a.__pformat(indentation+2)
        
        if oper in ['&', '|']:
            subs = []
            for e in self.terms:
                subs.append(e.__pformat(indentation+2))
            return oper+' '+('\n'+space*(indentation+2)).join(subs)
        
        assert(False)

    def __str__(self):
        return self.__pformat()

    def get_op(self):
        return self.__dict__.get('oper', None)

    def get_terms(self):
        return [x for x in self.__dict__.get('terms', [None])]

    def size(self):
        if '_size' in self.__dict__:
            return self._size
        if self.atomic:
            self._size = 1
            return 1
        self._size = 1 + sum([x.size() for x in self.terms])
        return self._size
    
    def depth(self):
        if '_depth' in self.__dict__:
            return self._depth
        if self.is_literal():
            self._depth = 1
            return 1
        self._depth = 1 + max([x.depth() for x in self.terms])
        return self._depth

    def var_stats(self):
        #if '_var_stats' in self.__dict__:
        #    return self._var_stats

        counts = dict()
        #self._var_stats = counts

        if self.is_literal():
            if self.is_positive():
                if self not in counts:
                    counts[self] = [0, 0]
                counts[self][0] += 1
            else:
                pos = self.get_terms()[0]
                if pos not in counts:
                    counts[pos] = [0, 0]
                counts[pos][1] += 1
            return counts
        
        if self.oper == '~':
            for sub in self.terms:
                nc = sub.var_stats()
                for e in nc:
                    if e not in counts:
                        counts[e] = [0, 0]
                    counts[e][0] += nc[e][1]
                    counts[e][1] += nc[e][0]
            return counts
        
        def update(a, counts):
            for e in a:
                if e not in counts:
                    counts[e] = a[e]
                else:
                    counts[e][0] += a[e][0]
                    counts[e][1] += a[e][1]

        if self.oper in ['&', '|']:
            for sub in self.terms:
                update(sub.var_stats(), counts)
            return counts

        raise NameError("ERROR")




"""Formatter to create the Belief Change Format (BCF) representation
of an arbitrary EquibelGraph instance.
"""
#    Copyright (C) 2014-2015 by
#    Paul Vicol <pvicol@sfu.ca>
#    All rights reserved.
#    BSD license.

__all__ = ['convert_to_bcf', 'convert_formula_to_bcf']


TYPE_TEMPLATE    = "t {0}\n"
NODE_TEMPLATE    = "n {0}\n"
RANGE_TEMPLATE   = "n {0}..{1}\n"
EDGE_TEMPLATE    = "e {0} {1}\n"
ATOM_TEMPLATE    = "a {0}\n"
WEIGHT_TEMPLATE  = "w {0} {1} {2}\n"
FORMULA_TEMPLATE = "f {0} {1}\n"

ENTAILMENT_CONSTRAINT_TEMPLATE  = "m {0}\n"
CONSISTENCY_CONSTRAINT_TEMPLATE = "s {0}\n"

# This dictionary maps from the operators used in the proposition
# class to the operators used in BCF.
bcf_op_dict = {
    '&': '*',
    '|': '+',
    '~': '-'
}


def convert_to_bcf(G):
    bcf_str = ""

    if G.is_directed():
        bcf_str += TYPE_TEMPLATE.format("directed")
    else:
        bcf_str += TYPE_TEMPLATE.format("undirected")

    for node_id in G.nodes():
        bcf_str += NODE_TEMPLATE.format(node_id)

    for atom in G.atoms():
        bcf_str += ATOM_TEMPLATE.format(atom)

    for (from_node_id, to_node_id) in G.edges():
        bcf_str += EDGE_TEMPLATE.format(from_node_id, to_node_id)
        if not G.is_directed():
            bcf_str += EDGE_TEMPLATE.format(to_node_id, from_node_id)

    for node_id in G.nodes():
        for atom in G.atom_weights(node_id):
            weight = G.weight(node_id, atom)
            bcf_str += WEIGHT_TEMPLATE.format(node_id, atom, weight)

    # This is separated from the above for loop for prettiness, to group
    # all the formulas together.
    for node_id in G.nodes():
        for formula in G.formulas(node_id):
            formatted_formula = "({0})".format(convert_formula_to_bcf(formula))
            bcf_str += FORMULA_TEMPLATE.format(node_id, formatted_formula)

    return bcf_str


def convert_formula_to_bcf(formula):
    if formula.is_atomic():
        return formula.get_name()

    terms = formula.get_terms()
    bcf_op = bcf_op_dict[formula.get_op()]
    return "{0}({1})".format(bcf_op,
                             " ".join([convert_formula_to_bcf(term) for term in terms]))



"""Formatter to create the Answer Set Programming (ASP) representation
of an arbitrary EquibelGraph instance. The ASP format is used when 
giving a graph to the gringo/clingo ASP tools.
"""
#    Copyright (C) 2014-2015 by
#    Paul Vicol <pvicol@sfu.ca>
#    All rights reserved.
#    BSD license.

from equibel.simbool.proposition import Prop


__all__ = ['convert_to_asp', 'convert_formula_to_asp']


NODE_TEMPLATE    = "node({0}).\n"
RANGE_TEMPLATE   = "node({0}..{1}).\n"
EDGE_TEMPLATE    = "edge({0},{1}).\n"
ATOM_TEMPLATE    = "atom({0}).\n"
WEIGHT_TEMPLATE  = "weight({0},{1},{2}).\n"
FORMULA_TEMPLATE = "formula({0},{1}).\n"

AND_TEMPLATE   = "and({0},{1})"
OR_TEMPLATE    = "or({0},{1})"
NEG_TEMPLATE   = "neg({0})"


def convert_to_asp(G):
    asp_str = ""

    for node_id in G.nodes():
        asp_str += NODE_TEMPLATE.format(node_id)

    for atom in G.atoms():
        asp_str += ATOM_TEMPLATE.format(atom)

    # DONE: This is one way to handle undirected edges: when writing the ASP
    #       code, explicitly encode both directions if the G is undirected.
    for (from_node_id, to_node_id) in G.edges():
        asp_str += EDGE_TEMPLATE.format(from_node_id, to_node_id)
        if not G.is_directed():
            asp_str += EDGE_TEMPLATE.format(to_node_id, from_node_id)

    for node_id in G.nodes():
        weights = G.atom_weights(node_id)
        for atom in weights:
            weight = weights[atom]
            asp_str += WEIGHT_TEMPLATE.format(node_id, atom, weight)

    # This is separated from the above for loop for prettiness, to group
    # all the formulas together.
    for node_id in G.nodes():
        formulas = G.formulas(node_id)
        for formula in formulas:
            formatted_formula = convert_formula_to_asp(formula)
            asp_str += FORMULA_TEMPLATE.format(formatted_formula, node_id)

    return asp_str


def convert_formula_to_asp(formula):
    # Atomic propositions are the base case for the recursion.
    if formula.is_atomic():
        name = formula.get_name()
        if name is True:
            return 'true'
        elif name is False:
            return 'false'
        else:
            return name

    if formula.get_op() == '~':
        term = formula.get_terms()[0]
        formatted_term = convert_formula_to_asp(term)
        return NEG_TEMPLATE.format(formatted_term)

    terms = formula.get_terms()
    if len(terms) == 1:
        # This handles the case when we have a conjunction/disjunction with
        # only one operand, like *(p) or +(p).
        term = formula.get_terms()[0]
        return convert_formula_to_asp(term)
    elif len(terms) == 2:
        first_operand = convert_formula_to_asp(terms[0])
        second_operand = convert_formula_to_asp(terms[1])
    else:
        first_operand = convert_formula_to_asp(terms[0])
        # This creates a new formula with the same operator as the one being
        # parsed, creating a smaller disjunction/conjunction (that is, one
        # with fewer operands). This is done so that recursive calls to this
        # function will produce binary formulas.
        rest_of_formula = Prop(formula.get_op(), *terms[1:])
        second_operand = convert_formula_to_asp(rest_of_formula)

    if formula.get_op() == '&':
        return AND_TEMPLATE.format(first_operand, second_operand)
    elif formula.get_op() == '|':
        return OR_TEMPLATE.format(first_operand, second_operand)



# lextab.py. This file automatically created by PLY (version 3.6). Don't edit!
_tabversion   = '3.5'
_lextokens    = set(['LSQUARE', 'RSQUARE', 'RPAREN', 'STRING', 'COMMA', 'LPAREN', 'INTEGER', 'DOT'])
_lexreflags   = 0
_lexliterals  = ''
_lexstateinfo = {'INITIAL': 'inclusive'}
_lexstatere   = {'INITIAL': [('(?P<t_INTEGER>0|[1-9][0-9]*)|(?P<t_NEWLINE>\\n+)|(?P<t_STRING>\\w+)|(?P<t_LPAREN>\\()|(?P<t_RSQUARE>\\])|(?P<t_DOT>\\.)|(?P<t_RPAREN>\\))|(?P<t_LSQUARE>\\[)|(?P<t_COMMA>,)', [None, ('t_INTEGER', 'INTEGER'), ('t_NEWLINE', 'NEWLINE'), (None, 'STRING'), (None, 'LPAREN'), (None, 'RSQUARE'), (None, 'DOT'), (None, 'RPAREN'), (None, 'LSQUARE'), (None, 'COMMA')])]}
_lexstateignore = {'INITIAL': ' \t\n'}
_lexstateerrorf = {'INITIAL': 't_error'}
_lexstateeoff = {}



"""The graph class used to represent networks of agents in Equibel.
"""
#    Copyright (C) 2014-2015 by
#    Paul Vicol <pvicol@sfu.ca>
#    All rights reserved.
#    MIT license.
from __future__ import absolute_import
from __future__ import print_function

import networkx as nx
import equibel

ATOMS_KEY    = "atoms"
WEIGHTS_KEY  = "weights"
FORMULAS_KEY = "formulas"

class EquibelGraph:

    def __init__(self, G=None):
        if G:
            self.graph = G.copy()
            self.graph.graph[ATOMS_KEY] = set()
            for node_id in G.nodes():
                self.graph.node[node_id][WEIGHTS_KEY] = dict()
                self.graph.node[node_id][FORMULAS_KEY] = set()
        else:
            self.graph = nx.Graph()
            self.graph.graph[ATOMS_KEY] = set()

    def __iter__(self):
        return iter(self.graph.nodes())

    def __getitem__(self, value):
        return self.graph[value]

    # ================================================
    #                NODE METHODS
    # ================================================

    # ROUTING to G.nodes()
    def nodes(self):
        return self.graph.nodes()

    def add_node(self, node_id):
        self.graph.add_node(node_id)
        self.graph.node[node_id][WEIGHTS_KEY]  = dict()
        self.graph.node[node_id][FORMULAS_KEY] = set()
        self.__add_atoms_to_new_node(node_id)

    def add_nodes(self, nodes):
        for node in nodes:
            self.add_node(node)

    def remove_node(self, node_id):
        self.graph.remove_node(node_id)

    # Perhaps this could be called "remove_nodes"
    def remove_nodes_from(self, nodes):
        self.graph.remove_nodes_from(nodes)

    # ================================================
    #                 EDGE METHODS
    # ================================================

    # ROUTING to G.edges()
    def edges(self):
        return self.graph.edges()

    def add_edge(self, from_node_id, to_node_id):
        if from_node_id not in self.graph:
            self.add_node(from_node_id)
        if to_node_id not in self.graph:
            self.add_node(to_node_id)

        self.graph.add_edge(from_node_id, to_node_id)

    def add_edges(self, edges):
        for (from_node_id, to_node_id) in edges:
            self.add_edge(from_node_id, to_node_id)

    def remove_edge(self, from_node_id, to_node_id):
        self.graph.remove_edge(from_node_id, to_node_id)

    # (not quite) ROUTING to G.to_directed()
    def to_directed(self):
        return EquibelGraph(self.graph.to_directed())

    # (not quite) ROUTING to G.to_undirected()
    def to_undirected(self):
        return EquibelGraph(self.graph.to_undirected())

    # ROUTING to G.is_directed()
    def is_directed(self):
        return self.graph.is_directed()

    def reverse(self):
        return EquibelGraph(self.graph.reverse())

    # ================================================
    #             ATOM and WEIGHT METHODS
    # ================================================

    # Instead of G.graph[ATOMS_KEY], use G.atoms()
    def atoms(self):
        return self.graph.graph[ATOMS_KEY]

    # Instead of G.node[node_id][WEIGHTS_KEY], use G.atom_weights(node_id)
    def atom_weights(self, node_id):
        return self.graph.node[node_id][WEIGHTS_KEY]

    def weight(self, node_id, atom):
        return self.graph.node[node_id][WEIGHTS_KEY][atom]

    def add_atom(self, atom):
        self.graph.graph[ATOMS_KEY].add(atom)
        self.__add_new_atom_to_nodes(atom)

    def add_atoms(self, atoms):
        for atom in atoms:
            self.add_atom(atom)

    def __add_new_atom_to_nodes(self, atom):
        for node_id in self.nodes():
            if atom not in self.atom_weights(node_id):
                self.set_atom_weight(node_id, atom, 1)

    def __add_atoms_to_new_node(self, node_id):
        for atom in self.atoms():
            self.set_atom_weight(node_id, atom, 1)

    def set_atom_weight(self, node_id, atom, weight):
        self.graph.node[node_id][WEIGHTS_KEY][atom] = weight
        if atom not in self.atoms():
            self.add_atom(atom)

    # For convenience:
    def add_weight(self, node_id, atom, weight):
        self.set_atom_weight(node_id, atom, weight)

    def remove_atom(self, atom):
        self.graph.graph[ATOMS_KEY].discard(atom)
        self.__remove_atom_from_nodes(atom)

    def __remove_atom_from_nodes(self, atom):
        for node_id in self.nodes():
            weights = self.atom_weights(node_id)
            if atom in weights:
                del weights[atom]

    # ================================================
    #               FORMULA METHODS
    # ================================================

    # Instead of G.node[node_id][FORMULAS_KEY], use G.formulas(node_id)
    def formulas(self, node_id):
        return self.graph.node[node_id][FORMULAS_KEY]

    def add_formula(self, node_id, formula):
        if isinstance(formula, str):
            parsed_formula = equibel.parse_infix_formula(formula)
            self.graph.node[node_id][FORMULAS_KEY].add(parsed_formula)
            self.__add_formula_atoms(parsed_formula)
        else:
            self.graph.node[node_id][FORMULAS_KEY].add(formula)
            self.__add_formula_atoms(formula)

    def __add_formula_atoms(self, formula):
        for atom in formula.get_atoms():
            if atom not in self.atoms():
                self.add_atom(atom)

    def set_formulas(self, node_id, formula_list):
        self.clear_formulas_from(node_id)
        for formula in formula_list:
            self.graph.node[node_id][FORMULAS_KEY].add(formula)

    def clear_formulas_from(self, node_id):
        self.graph.node[node_id][FORMULAS_KEY].clear()

    def clear_formulas(self):
        for node_id in self.nodes():
            self.graph.node[node_id][FORMULAS_KEY].clear()

    def remove_formula(self, node_id, formula):
        self.graph.node[node_id][FORMULAS_KEY].discard(formula)




# /Users/paulvicol/Code/Python/equibel_install/equibel/PredicateTree/parsetab.py
# This file is automatically generated. Do not edit.
_tabversion = '3.5'

_lr_method = 'LALR'

_lr_signature = '478A82D5E30F8251E282BA9BF3BAF075'
    
_lr_action_items = {'RPAREN':([4,5,6,7,8,9,11,],[-6,9,-2,-4,-5,-1,-3,]),'COMMA':([4,6,7,8,9,],[-6,10,-4,-5,-1,]),'LPAREN':([2,8,],[3,3,]),'INTEGER':([3,10,],[7,7,]),'IDENTIFIER':([0,3,10,],[2,8,8,]),'$end':([1,9,],[0,-1,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'PREDICATE':([0,3,10,],[1,4,4,]),'ARGS':([3,10,],[5,11,]),'ARG':([3,10,],[6,6,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> PREDICATE","S'",1,None,None,None),
  ('PREDICATE -> IDENTIFIER LPAREN ARGS RPAREN','PREDICATE',4,'p_PREDICATE','predicate_parser.py',40),
  ('ARGS -> ARG','ARGS',1,'p_ARGS','predicate_parser.py',45),
  ('ARGS -> ARG COMMA ARGS','ARGS',3,'p_ARGS','predicate_parser.py',46),
  ('ARG -> INTEGER','ARG',1,'p_ARG','predicate_parser.py',51),
  ('ARG -> IDENTIFIER','ARG',1,'p_ARG','predicate_parser.py',52),
  ('ARG -> PREDICATE','ARG',1,'p_ARG','predicate_parser.py',53),
]



from __future__ import absolute_import

import equibel.graph
from equibel.graph import *

import equibel.generator
from equibel.generator import *

import equibel.graph_manager
from equibel.graph_manager import *

import equibel.simbool
from equibel.simbool import *

import equibel.formatters
from equibel.formatters import *

import equibel.parsers
from equibel.parsers import *

import equibel.parsers.BCF_Parser
from equibel.parsers.BCF_Parser import *

import equibel.solver
from equibel.solver import *



from __future__ import absolute_import

import sys

from ply import lex
from ply import yacc

from equibel.PredicateTree.predicate import Predicate

tokens = ("IDENTIFIER", "INTEGER", "LPAREN", "RPAREN", "COMMA")

t_IDENTIFIER = r'[_a-zA-Z][_a-zA-Z0-9]*'
t_LPAREN = r'\('
t_RPAREN = r'\)'
t_COMMA  = r','

t_ignore = " \t\n"


def t_INTEGER(t):
    r"0|[1-9][0-9]*"
    t.value = int(t.value)
    return t


def t_NEWLINE(t):
    r"\n+"
    t.lexer.lineno += len(t.value)
    return t


def t_error(t):
    line = t.value.lstrip()
    i = line.find("\n")
    line = line if i == -1 else line[:i]
    raise ValueError("Syntax error, line {0}: {1}".format(t.lineno + 1, line))


def p_PREDICATE(p):
    """PREDICATE : IDENTIFIER LPAREN ARGS RPAREN"""
    p[0] = Predicate(p[1], p[3])


def p_ARGS(p):
    """ARGS : ARG
           | ARG COMMA ARGS"""
    p[0] = [p[1]] if len(p) == 2 else [p[1]] + p[3]


def p_ARG(p):
    """ARG : INTEGER
          | IDENTIFIER
          | PREDICATE"""
    p[0] = p[1]


def p_error(p):
    if p is None:
        raise ValueError("Unknown error")
    raise ValueError("Syntax error, line {0}: {1}".format(p.lineno + 1, p.type))

lexer = lex.lex()
parser = yacc.yacc()


def parse_predicate(text):
    try:
        return parser.parse(text, lexer=lexer)
    except ValueError as err:
        print(err)


if __name__ == '__main__':
    if len(sys.argv) < 2:
        print('usage: python3 PredicateParser.py filename')
        sys.exit(1)

    filename = sys.argv[1]
    f = open(filename, 'r')

    try:
        for line in f.readlines():
            predicate = parse_predicate(line)
            print(predicate)
    except Exception as err:
        print(err)



#    Copyright (C) 2014-2015 by
#    Paul Vicol <pvicol@sfu.ca>
#    All rights reserved.
#    BSD license.

from __future__ import absolute_import

import sys

import ply.lex
import ply.yacc

from equibel.simbool.proposition import *
from equibel.simbool.simplify import *


__all__ = ["parse_infix_formula"]


# --------------------------------------------------------------------
#                             LEXER
# --------------------------------------------------------------------
keywords = {"T": "TRUE", "F": "FALSE"}

tokens = (["NEG", "AND", "OR", "IMPLIES", "EQUIV", "LPAREN", "RPAREN", "IDENTIFIER"] + list(keywords.values()))

def t_IDENTIFIER(t):
    r"[_a-zA-Z][_a-zA-Z0-9]*"
    t.type = keywords.get(t.value, "IDENTIFIER")
    return t

t_NEG = r"~"
t_AND = r"&"
t_OR = r"\|"
t_IMPLIES = r"->"
t_EQUIV = r"="
t_LPAREN = r"\("
t_RPAREN = r"\)"

t_ignore = " \t\n"

def t_NEWLINE(t):
    r"\n+"
    t.lexer.lineno += len(t.value)
    return t

def t_error(t):
    line = t.value.lstrip()
    i = line.find("\n")
    line = line if i == -1 else line[:i]
    raise ValueError("Syntax error, line {0}: {1}".format(t.lineno + 1, line))


# --------------------------------------------------------------------
#                             PARSER
# --------------------------------------------------------------------

def p_FORMULA(p):
    """FORMULA : ATOM
               | BOOLEAN
               | COMPOUND
               | LPAREN FORMULA RPAREN"""
    p[0] = p[1] if len(p) == 2 else p[2]

def p_ATOM(p):
    """ATOM : IDENTIFIER"""
    p[0] = Prop(p[1])

def p_BOOLEAN_TRUE(p):
    """BOOLEAN : TRUE"""
    p[0] = Prop(True)

def p_BOOLEAN_FALSE(p):
    """BOOLEAN : FALSE"""
    p[0] = Prop(False)

def p_COMPOUND(p):
    """COMPOUND : NEGATION
                | CONJUNCTION
                | DISJUNCTION
                | IMPLICATION
                | EQUIVALENCE"""
    p[0] = p[1]

def p_NEGATION(p):
    """NEGATION : NEG FORMULA"""
    p[0] = ~ p[2]

def p_CONJUNCTION(p):
    """CONJUNCTION : FORMULA AND FORMULA"""
    p[0] = p[1] & p[3]

def p_DISJUNCTION(p):
    """DISJUNCTION : FORMULA OR FORMULA"""
    p[0] = p[1] | p[3]

def p_IMPLICATION(p):
    """IMPLICATION : FORMULA IMPLIES FORMULA"""
    # Material implication using operator ">"
    p[0] = p[1] > p[3]

def p_EQUIVALENCE(p):
    """EQUIVALENCE : FORMULA EQUIV FORMULA"""
    p[0] = (p[1] > p[3]) & (p[3] > p[1])

def p_error(p):
    if p is None:
        raise ValueError("Unknown error")
    raise ValueError("Syntax error, line {0}: {1}".format(p.lineno + 1, p.type))


precedence = (("right", "EQUIV"),
              ("right", "IMPLIES"),
              ("left", "OR"),
              ("left", "AND"),
              ("right", "NEG"))


lexer = ply.lex.lex(debug=False)
parser = ply.yacc.yacc(debug=False, write_tables=False)


def parse_infix_formula(text):
    try:
        return parser.parse(text, lexer=lexer)
    except ValueError as err:
        print(err)
        return []


if __name__ == '__main__':
    if len(sys.argv) < 2:
        print('usage: python3 formula_parser_infix.py file')
        sys.exit(1)

    formula_str = sys.argv[1]
    
    formula = parse_infix_formula(formula_str)
    
    print(repr(formula))
    print("Simplified = \n{0}".format(repr(simplify(formula))))



from setuptools import setup

setup(
    name='equibel',
    version='0.8.8a1', # v0.8.8 alpha build
    
    description='A toolkit for equivalence-based belief change',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
    ],

    keywords=['AI', 'belief change', 'belief merging', 'agent network'],

    url='https://github.com/asteroidhouse/equibel',
    author='Paul Vicol',
    author_email='pvicol@sfu.ca',
    license='MIT',

    packages=[
        'equibel',
        'equibel.simbool',
        'equibel.PredicateTree',
        'equibel.parsers',
        'equibel.formatters',
    ],

    include_package_data = True,

    install_requires=[
        'networkx',
        'ply',
        'colorama',
    ],

    entry_points='''
        [console_scripts]
        equibeli=equibel.equibeli:cli
    ''',
)



#    Copyright (C) 2014-2015 by
#    Paul Vicol <pvicol@sfu.ca>
#    All rights reserved.
#    BSD license.

from __future__ import absolute_import

import pprint
import sys

import ply.lex
import ply.yacc

from equibel.simbool.proposition import *
from equibel.simbool.simplify import *


__all__ = ["parse_formula"]


tokens = ("NEG", "AND", "OR", "LPAREN", "RPAREN", "IDENTIFIER", "WHITESPACE")

t_NEG = r'\-'
t_AND = r'\*'
t_OR = r'\+'
t_LPAREN = r'\('
t_RPAREN = r'\)'
t_IDENTIFIER = r'[_a-zA-Z][_a-zA-Z0-9]*'

t_WHITESPACE = r'[ \t]+'

def t_NEWLINE(t):
    r"\n+"
    t.lexer.lineno += len(t.value)
    return t

def t_error(t):
    line = t.value.lstrip()
    i = line.find("\n")
    line = line if i == -1 else line[:i]
    raise ValueError("Syntax error, line {0}: {1}".format(t.lineno + 1, line))

def p_FORMULA_LINE(p):
    """FORMULA_LINE : OPT_WHITESPACE FORMULA OPT_WHITESPACE"""
    p[0] = p[2]

def p_FORMULA(p):
    """FORMULA : LPAREN OPT_WHITESPACE FORM OPT_WHITESPACE RPAREN"""
    #print("formula")
    p[0] = p[3]

def p_FORM(p):
    """FORM : SIMPLE_FORM
            | COMPOUND_FORM"""
    #print("form")
    p[0] = p[1]

def p_SIMPLE_FORM(p):
    """SIMPLE_FORM : ATOM
                   | NEG OPT_WHITESPACE ATOM
                   | LPAREN OPT_WHITESPACE FORM OPT_WHITESPACE RPAREN
                   | NEG OPT_WHITESPACE LPAREN OPT_WHITESPACE FORM OPT_WHITESPACE RPAREN"""
    #print("simple form")
    if len(p) == 2:
        p[0] = p[1]
    elif len(p) == 4:
        p[0] = Prop('~', p[3])
    elif len(p) == 6:
        p[0] = p[3]
    elif len(p) == 8:
        p[0] = Prop('~', p[5])

def p_COMPOUND_FORM(p):
    """COMPOUND_FORM : OP OPT_WHITESPACE LPAREN OPT_WHITESPACE FORM_LIST OPT_WHITESPACE RPAREN"""
    p[0] = Prop(p[1], *p[5])

def p_FORM_LIST(p):
    """FORM_LIST : FORM OPT_WHITESPACE
                 | FORM WHITESPACE FORM_LIST"""
    #print("form list")
    p[0] = [p[1]] if len(p) == 3 else [p[1]] + p[3]

def p_ATOM(p):
    """ATOM : IDENTIFIER"""
    #print("atom")
    p[0] = Prop(p[1])

def p_OP(p):
    """OP : AND_OP
          | OR_OP"""
    #print("op")
    p[0] = p[1]

# This rule was introduced to translate between the notation used 
# to represent a conjunction in the input language and the notation
# used by the simbool module.
def p_AND_OP(p):
    """AND_OP : AND"""
    p[0] = '&'

# This rule was introduced to translate between the notation used 
# to represent a disjunction in the input language and the notation
# used by the simbool module.
def p_OR_OP(p):
    """OR_OP : OR"""
    p[0] = '|'

def p_OPT_WHITESPACE(p):
    """OPT_WHITESPACE : WHITESPACE
                      | empty"""
    pass

def p_empty(p):
    """empty :"""
    pass

def p_error(p):
    if p is None:
        raise ValueError("Unknown error")
    raise ValueError("Syntax error, line {0}: {1}".format(p.lineno + 1, p.type))

lexer = ply.lex.lex(debug=False)
parser = ply.yacc.yacc(debug=False, write_tables=False)

def parse_formula(text):
    try:
        return parser.parse(text, lexer=lexer)
    except ValueError as err:
        print(err)
        return []


if __name__ == '__main__':
    if len(sys.argv) < 2:
        print('usage: python3 FormulaParserSim.py file')
        sys.exit(1)

    filename = sys.argv[1]
    f = open(filename, 'r')
    
    formula = parse_formula(f.read().strip())
    
    print(repr(formula))
    print("Simplified = \n{0}".format(repr(simplify(formula))))



#    Copyright (C) 2014-2015 by
#    Paul Vicol <pvicol@sfu.ca>
#    All rights reserved.
#    BSD license.

from ply import lex
from ply import yacc


tokens = ["STRING", "INTEGER", "COMMA", "DOT", "LPAREN", "RPAREN", "LSQUARE", "RSQUARE"]

t_STRING  = r'\w+'
t_COMMA   = r','
t_DOT     = r'\.'
t_LPAREN  = r'\('
t_RPAREN  = r'\)'
t_LSQUARE = r'\['
t_RSQUARE = r'\]'

# Ignore whitespace
t_ignore = " \t\n"


def t_INTEGER(t):
    r"0|[1-9][0-9]*"
    t.value = int(t.value)
    return t

def t_NEWLINE(t):
    r"\n+"
    t.lexer.lineno += len(t.value)
    return t

def t_error(t):
    line = t.value.lstrip()
    i = line.find("\n")
    line = line if i == -1 else line[:i]
    raise ValueError("Syntax error, line {0}: {1}".format(t.lineno + 1, line))

def p_TUPLE(p):
    """TUPLE : LPAREN INTEGER COMMA INTEGER RPAREN"""
    p[0] = (int(p[2]), int(p[4]))

def p_error(p):
    if p is None:
        raise ValueError("Unknown error")
    raise ValueError("Syntax error, line {0}: {1}".format(p.lineno + 1, p.type))


def parse_tuple(text):
    lexer = lex.lex(optimize=1, debug=False)
    tuple_parser = yacc.yacc(start='TUPLE', debug=False, write_tables=False)

    try:
        return tuple_parser.parse(text, lexer=lexer)
    except ValueError as err:
        print(err)


def parse_list(text):

    def p_LIST(p):
        """LIST : LSQUARE ITEMS RSQUARE
                | LSQUARE RANGE RSQUARE
                | LSQUARE RSQUARE"""
        p[0] = [] if len(p) == 3 else p[2]

    def p_RANGE(p):
        """RANGE : INTEGER DOT DOT INTEGER"""
        p[0] = list(range(p[1], p[4] + 1))

    def p_ITEMS(p):
        """ITEMS : ITEM
                 | ITEM COMMA ITEMS"""
        p[0] = [p[1]] if len(p) == 2 else [p[1]] + p[3]

    def p_ITEM(p):
        """ITEM : INTEGER
                | STRING
                | TUPLE"""
        p[0] = p[1]

    lexer = lex.lex(optimize=1, debug=False)
    list_parser  = yacc.yacc(start='LIST', debug=False, write_tables=False)

    try:
        return list_parser.parse(text, lexer=lexer)
    except ValueError as err:
        print(err)



"""Equibel wrappers for networkx graph generator functions.

This file contains wrapper functions for several networkx graph 
generators. The wrappers return EquibelGraph objects that contain 
the corresponding networkx graphs.
"""
#    Copyright (C) 2014-2015 by
#    Paul Vicol <pvicol@sfu.ca>
#    All rights reserved.
#    BSD license.
from __future__ import absolute_import
from __future__ import print_function

import networkx as nx
import sys

from equibel.graph import EquibelGraph


# ======================================================
# Classic Graph Generators -- Line, Star, Complete, etc.
# ======================================================

def path_graph(n, directed=False):
    """Create a directed or undirected line graph on n nodes, 
    numbered using the integers 1 to n, inclusive.

    Parameters
    ----------
    n        : int
               The desired number of nodes for the line graph.
    directed : boolean
               If set to True, then the edges of the line graph 
               will be directed in order of increasing node number.

    Returns
    -------
    G : EquibelGraph
        An EquibelGraph representing a line graph.

    Examples
    --------
    Create an undirected line graph 1 <-> 2 <-> 3 <-> 4:
    >>> G = line_graph(4)

    Create a directed line graph 1 -> 2 -> 3 -> 4:
    >>> D = line_graph(4, directed=True)
    """
    if directed:
        return EquibelGraph(nx.path_graph(n, create_using=nx.DiGraph()))
    else:
        return EquibelGraph(nx.path_graph(n))

def complete_graph(n):
    """Create an undirected complete graph on n nodes.

    Parameters
    ----------
    n : int
        The desired number of nodes for the complete graph.

    Returns
    -------
    G : EquibelGraph
        An EquibelGraph containing a complete graph on n nodes.

    Examples
    --------
    >>> G = complete_graph(5)
    """
    return EquibelGraph(nx.complete_graph(n))

def star_graph(n):
    """Create an undirected star graph on n+1 nodes, with one 
    central node, and n outer nodes.
    """
    return EquibelGraph(nx.star_graph(n))


# ========================================================
# Random Graph Generators -- Waxman, Barabasi-Albert, etc.
# ========================================================

def waxman_graph(n, alpha=0.4, beta=0.1, L=None, domain=(0,0,1,1)):
    """
    """
    return EquibelGraph(nx.waxman_graph(n, alpha, beta, L, domain))

def erdos_renyi_graph(n, p, seed=None, directed=False):
    """
    """
    return EquibelGraph(nx.erdos_renyi_graph(n, p, seed, directed))



#    Copyright (C) 2014-2015 by
#    Paul Vicol <pvicol@sfu.ca>
#    All rights reserved.
#    BSD license.

from __future__ import absolute_import
from __future__ import print_function

import pkg_resources
import copy
import platform

import equibel

if platform.system() == 'Linux':
    if platform.architecture()[0] == '64bit':
        import equibel.includes.linux.bit64.gringo as gringo
        from equibel.includes.linux.bit64.gringo import Control, Model, Fun
    elif platform.architecture()[0] == '32bit':
        import equibel.includes.linux.bit32.gringo as gringo
        from equibel.includes.linux.bit32.gringo import Control, Model, Fun
elif platform.system() == 'Darwin':
    import equibel.includes.mac.gringo as gringo
    from equibel.includes.mac.gringo import Control, Model, Fun

import equibel.FormulaExtractor as FormulaExtractor
import equibel.formatters.ASP_Formatter as ASP_Formatter
from equibel.simbool.proposition import Prop
from equibel.simbool.simplify import simplify


CONTAINMENT = 'containment'
CARDINALITY = 'cardinality'

EQ_SETS_FILE = pkg_resources.resource_filename('equibel', 'asp/eq_sets.lp')
TRANSITIVE_FILE = pkg_resources.resource_filename('equibel', 'asp/transitive.lp')
TRANSLATE_FILE = pkg_resources.resource_filename('equibel', 'asp/translate.lp')
EQ_EXPANDING_FILE = pkg_resources.resource_filename('equibel', 'asp/eq_expanding.lp')


class EqSolver:

    def __init__(self):
        self.optimal_models = set()
        self.optimal_values = None

    def find_models(self, asp_string, method=CONTAINMENT):
        self.optimal_models = set()
        self.optimal_values = None

        ctl = gringo.Control()

        ctl.load(EQ_SETS_FILE)
        ctl.load(TRANSITIVE_FILE)
        ctl.load(TRANSLATE_FILE)

        self._configure_control(ctl, method)

        ctl.add('base', [], asp_string)
        ctl.ground([('base', [])])
        ctl.solve(on_model=self.capture_optimal_models)

        print("SOLVER MODELS = {0}".format(self.optimal_models))

        return self.optimal_models

    """
    def iteration(self, asp_string, method=CONTAINMENT):
        self.optimal_models = set()
        self.optimal_values = None

        ctl = gringo.Control()
        
        ctl.load(EQ_SETS_FILE)
        ctl.load(TRANSLATE_FILE)

        self._configure_control(ctl, method)

        ctl.add('base', [], asp_string)
        ctl.ground([('base', [])])
        ctl.solve(on_model=self.capture_optimal_models)

        return self.optimal_models
    
    def expanding_iteration(self, asp_string, method=CONTAINMENT):
        self.optimal_models = set()
        self.optimal_values = None

        # 1. Update formulas for each node based on the formulas 
        #    at nodes 1 hop away.
        ctl = gringo.Control()

        ctl.load(EQ_EXPANDING_FILE)
        ctl.load(TRANSLATE_FILE)

        self._configure_control(ctl, method)

        ctl.add('base', [], asp_string)
        ctl.ground([('base', [])])
        ctl.solve(on_model=self.capture_optimal_models)

        return self.optimal_models
    """

    def one_shot_dicts(self, asp_string, method=CONTAINMENT):
        models = self.find_models(asp_string, method)
        return self.model_dicts(models)

    """
    def iteration_dicts(self, asp_string, method=CONTAINMENT):
        models = self.iteration(asp_string, method)
        return self.model_dicts(models)
    
    def expanding_iteration_dicts(self, asp_string, method=CONTAINMENT):
        models = self.expanding_iteration(asp_string, method)
        return self.model_dicts(models)
    """

    def model_dicts(self, models):
        model_list = []

        for model in models:
            term_dict = dict()
            term_names = set(term.name() for term in model)
            for term_name in term_names:
                term_dict[term_name] = [term for term in model if term.name() == term_name]
            model_list.append(term_dict)

        return model_list

    def _configure_control(self, ctl, method):
        if method == CONTAINMENT:
            #ctl.conf.solve.opt_mode   = 'ignore'
            ctl.conf.solve.enum_mode  = 'domRec'
            ctl.conf.solver.heuristic = 'domain'
            ctl.conf.solve.models = 0
        elif method == CARDINALITY:
            ctl.conf.solve.opt_mode   = 'optN'
            ctl.conf.solve.enum_mode  = 'auto'
            ctl.conf.solver.heuristic = 'none'

    def capture_optimal_models(self, model):
        # model_atoms = frozenset(atom for atom in model.atoms(Model.SHOWN) if atom.name() == 'new_formula')
        model_atoms = frozenset(model.atoms(Model.SHOWN))
        opt_values = model.optimization()

        if self.optimal_values is None:
            self.optimal_values = opt_values
            self.optimal_models.add(model_atoms)
        elif opt_values == self.optimal_values:
            self.optimal_models.add(model_atoms)
        elif opt_values < self.optimal_values:
            self.optimal_values = opt_values
            self.optimal_models.clear()
            self.optimal_models.add(model_atoms)


# TEST TEST TEST
def completion(G, solving_method=CONTAINMENT):
    solver = EqSolver()
    models = solver.one_shot_dicts(ASP_Formatter.convert_to_asp(G), solving_method)
    #print("MODELS = {0}".format(models))
    node_formulas = FormulaExtractor.combine_formulas(models)

    R = copy.deepcopy(G)

    # TODO: Decide whether to always represent "multiple formulas" as a
    # single conjunction, or whether to keep them separate.
    for node_id in R.nodes():
        if node_id in node_formulas:
            new_formula = node_formulas[node_id]
            old_formula = conjunction(R.formulas(node_id))
            R.set_formulas(node_id, [simplify(old_formula & new_formula)])
            #R.set_formulas(node_id, [old_formula & new_formula])

    return R

# TEST TEST TEST
"""
def iterate(G, num_iterations=1, solving_method=CONTAINMENT):
    solver = EqSolver()
    asp_string = ASP_Formatter.convert_to_asp(G)

    node_formulas = None

    for i in range(num_iterations):
        models = solver.iteration_dicts(asp_string)
        print(models)
        node_formulas = FormulaExtractor.combine_formulas(models)
        for node_id in node_formulas:
            formula = node_formulas[node_id]
            asp_string += "formula({formula}, {node}).".format(formula=formula, node=node_id)

    R = copy.deepcopy(G)

    # TODO: Decide whether to always represent "multiple formulas" as a 
    # single conjunction, or whether to keep them separate.
    for node_id in R.nodes():
        if node_id in node_formulas:
            new_formula = node_formulas[node_id]
            old_formula = conjunction(R.formulas(node_id))
            R.set_formulas(node_id, [simplify(old_formula & new_formula)])

    return R
"""

def conjunction(formulas):
    result = Prop(True)

    for formula in formulas:
        result &= formula

    return simplify(result)


# TEST TEMPORARY
"""
def expanding_iteration(G, num_iterations=1, solving_method=CONTAINMENT):
    solver = EqSolver()
    asp_string = ASP_Formatter.convert_to_asp(G)

    if G.is_directed():
        D = G.reverse()
        length = nx.all_pairs_shortest_path_length(D)
    else:
        length = nx.all_pairs_shortest_path_length(D)

    for from_node_id in length:
        dists = length[from_node_id]
        for to_node_id in dists:
            distance = dists[to_node_id]
            asp_string += "dist({0},{1},{2}).".format(from_node_id, to_node_id, distance)

    print(asp_string)
"""


# =================================================
# Functions that involve implicit graph topologies: 
#   - Revision
#   - Contraction
#   - Merging (Projection and Consensus Based)
#   - Extrapolation
# ================================================= 

def revise(K, R):
    G = equibel.path_graph(2)

    if isinstance(K, str):
        G.add_formula(0, equibel.parse_infix_formula(K))
    elif isinstance(K, list):
        for belief in K:
            G.add_formula(0, equibel.parse_infix_formula(belief))

    if isinstance(R, str):
        G.add_formula(1, equibel.parse_infix_formula(R))
    elif isinstance(R, list):
        for belief in R:
            G.add_formula(1, equibel.parse_infix_formula(belief))

    S = completion(G)
    return S.formulas(1)


def contract(K, C):
    pass
    
def proj_merge(belief_bases, entailment_constraint=None):
    G = equibel.star_graph(len(belief_bases))
    
    if entailment_constraint:
        G.add_formula(0, entailment_constraint)

    for (i, belief_base) in enumerate(belief_bases, 1):
        G.add_formula(i, belief_base)

    S = completion(G)
    return S.formulas(0)
    

def con_merge(belief_bases, entailment_based_constraints=None, consistency_based_constraints=None):
    pass


if __name__ == '__main__':
    a = Prop('a')
    b = Prop('b')
    c = Prop('c')

    form1 = a & b
    form2 = b | ~c

    print(repr(conjunction([form1, form2])))

    solver = EqSolver()
    final_models2 = solver.find_models(open('TestCases/case8.lp').read(), method=EqSolver.CONTAINMENT)
    final_models = solver.model_dicts(open('TestCases/case9.lp').read(), method=EqSolver.CARDINALITY)
    
    for model in final_models:
        print(model['eq'])

    formulas = FormulaExtractor.combine_formulas(final_models)

    print(formulas)

    print("\n")

    for model in final_models2:
        print(model)

